import logging
import os
import platform
import sys

import sentry_sdk
from sentry_sdk import set_user
from torch.mps import empty_cache as mps_empty_cache

from spice_agent.__version__ import __version__
from spice_agent.auth.actions import Auth
from spice_agent.daemons.actions import Daemons
from spice_agent.graphql.sdk import create_session
from spice_agent.hardware.actions import Hardware
from spice_agent.inference.actions import Inference
from spice_agent.training.actions import Training
from spice_agent.uploader.actions import Uploader
from spice_agent.utils.config import read_config_file
from spice_agent.utils.version import update_if_outdated
from spice_agent.worker.actions import Worker

os.environ["PYTORCH_ENABLE_MPS_FALLBACK"] = "1"

import torch  # noqa


class Spice:
    def __init__(self, host: str = "api.spice.cloud", DEBUG: bool = False) -> None:
        self.host = host
        self.DEBUG = DEBUG

        logger = logging.getLogger("spice_agent")
        logger.setLevel(logging.INFO)
        sh = logging.StreamHandler(sys.stdout)
        formatter = logging.Formatter(
            "[%(asctime)s] %(levelname)s [%(filename)s.%(funcName)s:%(lineno)d] %(message)s",  # noqa
            datefmt="%a, %d %b %Y %H:%M:%S",
        )
        sh.setFormatter(formatter)
        logger.addHandler(sh)

        environment = "production"
        if "localhost" in host:
            environment = "development"
        if "staging" in host:
            environment = "staging"

        if environment != "development":
            update_if_outdated()

            # https://docs.sentry.io/product/sentry-basics/dsn-explainer/
            # It's okay to send the user's the Sentry DSN!
            # That's how we get metrics / errors from their usage.
            SENTRY_DSN = "https://1ee4a12126f0421bbe382d9227e46c4a@o4505155992223744.ingest.sentry.io/4505156109139968"  # noqa
            sentry_sdk.init(
                dsn=SENTRY_DSN,
                environment=environment,
                debug=DEBUG,
                traces_sample_rate=1.0,
                profiles_sample_rate=1.0,
                attach_stacktrace=True,
                send_default_pii=True,
                release=f"spice@{__version__}",
            )

        self.create_session()

        self.worker = Worker(self)
        self.auth = Auth(self)
        self.hardware = Hardware(self)
        self.inference = Inference(self)
        self.training = Training(self)
        self.uploader = Uploader(self)
        self.daemons = Daemons(self)

    def get_device(self):
        """
        First check if mps is available as a device
        Then check for a CUDA device
        Finally, fall back to CPU
        """
        device = None
        os_family = platform.system()

        # mps device enables high-performance training on GPU for macOS
        # devices with Metal programming framework
        # https://pytorch.org/docs/master/notes/mps.html
        if os_family == "Darwin" and torch.backends.mps.is_available():  # type: ignore
            device = torch.device("mps")
            mps_empty_cache()
            if self.DEBUG:
                print("Using MPS device.")
        else:
            if device is None and self.DEBUG:
                # in debug mode why is it not available
                if not torch.backends.mps.is_built():  # type: ignore
                    print(
                        "MPS not available because the current PyTorch install was not built with MPS enabled."  # noqa
                    )
                else:
                    print(
                        "MPS not available because the current macOS version is not 12.3+ and/or you do not have an MPS-enabled device on this machine."  # noqa
                    )

        if device is None and torch.cuda.is_available():
            device = torch.device("cuda:0")
            torch.cuda.empty_cache()
            if self.DEBUG:
                print("Using CUDA device.")
        else:
            if device is None and self.DEBUG:
                # in debug mode why is it not available
                if not torch.backends.cuda.is_built():  # type: ignore
                    print(
                        "CUDA not available because the current PyTorch install was not built with CUDA enabled."  # noqa
                    )
                else:
                    print(
                        "CUDA not available because the current you do not have an CUDA-enabled device on this machine."  # noqa
                    )

        if device is None:
            # fallback to CPU
            device = torch.device("cpu")
            if self.DEBUG:
                print("Using cpu.")

        return device

    def create_session(self):
        self.full_config = read_config_file()
        self.host_config = self.full_config.get(self.host)

        if not self.host_config:
            raise KeyError(f"Host {self.host} not found in config file.")
        else:
            set_user({"username": self.host_config["username"]})

        self.session = create_session(host=self.host, host_config=self.host_config)



import os
import webbrowser

import click

from spice_agent.auth.actions import Auth
from spice_agent.client import Spice

from ..utils.config import SPICE_HOSTS_FILEPATH
from ..utils.printer import print_result


@click.group()
@click.pass_context
def cli(context):
    """Authentication"""
    pass


@cli.command("whoami")
@click.pass_context
def whoami_command(context):
    """Whoami"""
    spice = context.obj.get("SPICE")
    result = spice.auth.whoami()
    message = f"Logged in as {result['whoami']['username']}"
    if context.obj["JSON"]:
        message = result
    print_result(message=message, context=context)


@cli.command("config")
@click.pass_context
@click.option(
    "--username",
    prompt=True,
    default=lambda: os.environ.get("SPICE_USER", ""),
    show_default="current user",
    help="Username for Spice API",
)
@click.option(
    "--transport",
    required=False,
    default="https",
    show_default="https",
    help="Transport protocol for Spice API",
)
@click.option(
    "--register",
    is_flag=True,
    show_default=True,
    default=False,
    help="Auto Register Hardware with Spice Cloud",
)
def config_command(context, username: str, transport: str, register: bool):
    """Configure the CLI"""
    token = os.environ.get("SPICE_TOKEN", "")
    if not token and context.obj.get("YES"):
        raise click.ClickException(
            "SPICE_TOKEN environment variable is required for non-interactive mode"
        )
    host = context.obj.get("HOST", "api.spice.cloud")
    while not token:
        settings_url = f"{transport}://{host.strip('api.')}/settings"
        if "localhost" in host:
            settings_url = "http://localhost:3000/settings"
        click.secho(
            f"You can find or create a Spice API token at {settings_url}",
            fg="yellow",
        )

        webbrowser.open_new_tab(settings_url)
        token = click.prompt(
            "Please enter your Spice API token", hide_input=True, type=str
        )

    auth = Auth(spice=None)
    auth.setup_config(username=username, token=token, host=host, transport=transport)
    message = f"Config created at '{SPICE_HOSTS_FILEPATH}' for user '{username}' on host '{host}'"  # noqa
    print_result(message=message, context=context, fg="green")

    if register or click.confirm(
        "Do you want to register this machine with spice.cloud?"
    ):
        spice = Spice(host=host)
        result = spice.hardware.register()

        if result.get("registerHardware"):
            print_result(
                message="Hardware registered successfully", context=context, fg="green"
            )

    if register or click.confirm(
        "Do you want to install the agent as a background process?"
    ):
        spice = Spice(host=host)
        spice.daemons.install()
        print_result(
            message="Daemon installed. View its logs with `spice daemon logs`",
            context=context,
            fg="green",
        )



# tests/test_version.py

import os
import toml
from spice_agent import __version__


def test_version():
    # Get the path to the pyproject.toml file
    project_dir = os.path.dirname(os.path.abspath("agent"))
    toml_file = os.path.join(project_dir, "pyproject.toml")

    # Read the version from pyproject.toml
    with open(toml_file) as f:
        config = toml.load(f)
        expected_version = config["tool"]["poetry"]["version"]

    # Ensure package version is the same as project version
    assert __version__.__version__ == expected_version



from typing import Tuple


class MemorySize:
    CONVERSION_FACTORS: dict[str, float] = {
        "b": 1 / 8,
        "B": 1,
        "KB": 1000,
        "MB": 1000**2,
        "GB": 1000**3,
        "TB": 1000**4,
        "PB": 1000**5,
        "KiB": 2**10,
        "MiB": 2**20,
        "GiB": 2**30,
        "TiB": 2**40,
        "PiB": 2**50,
    }

    def __init__(self, value_or_string: int | float | str, unit: str | None = None):
        """
        Initializes MemorySize in two ways:
            size = MemorySize("3.14 GB") - value_or_string is str and unit is None
            size = MemorySize(42, "MB") - value_or_string is int and unit is str
            size = MemorySize(2.718, "PB") - value_or_string is float and unit is str
        """
        if isinstance(value_or_string, str):
            parts = value_or_string.split()
            if len(parts) != 2:
                raise ValueError(f"Invalid size string format: {value_or_string}")

            value_string, unit = parts
            try:
                self.value = float(value_string)
            except ValueError:
                raise ValueError(f"Invalid size value: {value_string}")
            self.unit = unit
        elif isinstance(value_or_string, int | float) and isinstance(unit, str):
            self.value = float(value_or_string)
            self.unit = unit
        else:
            message = f"Invalid arguments provided: {value_or_string}:{type(value_or_string)} and {unit}:{type(unit)}"  # noqa
            raise ValueError(message)

        if not self.CONVERSION_FACTORS.get(self.unit):
            message = f"Invalid unit: {self.unit}"
            raise ValueError(message)

    def __str__(self):
        return f"{self.value:.3f} {self.unit}"

    def __eq__(self, other):
        if not isinstance(other, MemorySize):
            raise TypeError(f"Comparison with unsupported type: {type(other)}")
        return self.to_bytes() == other.to_bytes()

    def to_bytes(self):
        """
        Converts value to bytes (B)
        """
        return self.value * self.CONVERSION_FACTORS.get(self.unit, 1)

    def convert_to(self, target_unit):
        """
        Converts the stored value to a different unit and updates the instance.
        """
        if target_unit not in self.CONVERSION_FACTORS:
            raise ValueError(f"Invalid unit {target_unit}")

        factor = (
            self.CONVERSION_FACTORS[self.unit] / self.CONVERSION_FACTORS[target_unit]
        )
        self.value *= factor
        self.unit = target_unit

    def get_human_readable(self) -> Tuple[float, str]:
        """
        Converts the stored value to a human-readable format
        """
        HUMAN_READABLE_UNITS: list[str] = ["PB", "TB", "GB", "MB", "KB", "B"]

        value_in_bytes = self.to_bytes()
        for unit in HUMAN_READABLE_UNITS:
            value_in_unit = value_in_bytes * (1.000 / self.CONVERSION_FACTORS[unit])
            if value_in_unit >= 1.0:
                return (value_in_unit, unit)
        return (value_in_bytes, "B")



import platform

from spice_agent.daemons.launch_agents import (
    full_launch_agent_install,
    full_launch_agent_uninstall,
    start_launch_agent,
    stop_launch_agent,
    view_launch_agent_logs,
)

from spice_agent.daemons.launch_service import (
    full_service_install,
    full_service_uninstall,
    start_service,
    stop_service,
    view_service_logs,
)


class Daemons:
    def __init__(self, spice) -> None:
        self.spice = spice
        self.os_family = platform.system()

    def install(self):
        result = True
        if self.os_family == "Darwin":
            full_launch_agent_install()
        elif self.os_family == "Linux":
            full_service_install()
        elif self.os_family == "Windows":
            print("Not Implemented")
        return result

    def uninstall(self):
        result = True
        if self.os_family == "Darwin":
            full_launch_agent_uninstall()
        elif self.os_family == "Linux":
            full_service_uninstall()
        elif self.os_family == "Windows":
            print("Not Implemented")
        return result

    def stop(self) -> bool:
        result = True
        if self.os_family == "Darwin":
            result = stop_launch_agent()
        elif self.os_family == "Linux":
            stop_service()
        elif self.os_family == "Windows":
            print("Not Implemented")
        return result

    def start(self) -> bool:
        result = True
        if self.os_family == "Darwin":
            result = start_launch_agent()
        elif self.os_family == "Linux":
            start_service()
        elif self.os_family == "Windows":
            print("Not Implemented")
        return result

    def logs(self):
        if self.os_family == "Darwin":
            view_launch_agent_logs()
        elif self.os_family == "Linux":
            view_service_logs()
        elif self.os_family == "Windows":
            print("Not Implemented")



import csv
import io
import json
import logging
import platform
from shutil import which
import subprocess
from typing import Dict, List

from aiohttp import client_exceptions
import click
from gql import gql

from spice_agent.utils.config import (
    SPICE_TRAINING_FILEPATH,
    read_config_file,
    update_config_file,
)
from spice_agent.utils.memory_size import MemorySize

LOGGER = logging.getLogger(__name__)


class Hardware:
    def __init__(self, spice) -> None:
        self.spice = spice
        self.previous_state = {
            "isHealthy": False,
            "isQuarantined": False,
            "isAvailable": False,
            "isOnline": False,
        }

    def get_darwin_system_profiler_values(self) -> Dict[str, str]:
        system_profiler_hardware_data_type = subprocess.check_output(
            ["system_profiler", "SPHardwareDataType", "-json"]
        )
        system_profiler_hardware_data = json.loads(system_profiler_hardware_data_type)
        data_type = system_profiler_hardware_data.get("SPHardwareDataType")[0]
        return {
            "apple_model_name": data_type.get("machine_model"),
            "apple_model_identifier": data_type.get("machine_name"),
            "apple_model_number": data_type.get("model_number"),
            "physical_memory": data_type.get("physical_memory"),
            "apple_serial_number": data_type.get("serial_number"),
        }

    def _get_supported_metal_device(self) -> int | None:
        """
        Checks if metal hardware is supported. If so, the index
        of the supported metal device is returned
        """
        supported_metal_device = None
        is_system_profiler_available = bool(which("system_profiler"))
        if is_system_profiler_available:
            system_profiler_display_data_type_command = (
                "system_profiler SPDisplaysDataType -json"
            )
            try:
                system_profiler_display_data_type_output = subprocess.check_output(
                    system_profiler_display_data_type_command.split(" ")
                )
            except subprocess.CalledProcessError as exception:
                message = f"Error running system_profiler: {exception}"
                LOGGER.error(message)
                return supported_metal_device

            try:
                system_profiler_display_data_type_json = json.loads(
                    system_profiler_display_data_type_output
                )
            except json.JSONDecodeError as exception:
                message = f"Error decoding JSON: {exception}"
                LOGGER.error(message)
                return supported_metal_device

            # Checks if any attached displays have metal support
            # Note, other devices here could be AMD GPUs or unconfigured Nvidia GPUs
            for index, display in enumerate(
                system_profiler_display_data_type_json["SPDisplaysDataType"]
            ):
                if "spdisplays_mtlgpufamilysupport" in display:
                    supported_metal_device = index
                    return supported_metal_device

        return supported_metal_device

    def get_gpu_config(self) -> List:
        """
        For Nvidia based systems, nvidia-smi will be used to profile the gpu/s.
        For Metal based systems, we will gather information from SPDisplaysDataType.
        """
        gpu_config = []

        # Check nvidia gpu availability
        is_nvidia_smi_available = bool(which("nvidia-smi"))
        if is_nvidia_smi_available:
            nvidia_smi_query_gpu_csv_command = "nvidia-smi --query-gpu=gpu_name,driver_version,memory.total --format=csv"  # noqa
            try:
                nvidia_smi_query_gpu_csv_output = subprocess.check_output(
                    nvidia_smi_query_gpu_csv_command.split(" "),
                )
            except subprocess.CalledProcessError as exception:
                message = f"Command {nvidia_smi_query_gpu_csv_command} failed with exception: {exception}"  # noqa
                LOGGER.error(message)
                raise exception

            try:
                nvidia_smi_query_gpu_csv_decoded = (
                    nvidia_smi_query_gpu_csv_output.decode("utf-8")
                    .replace("\r", "")
                    .replace(", ", ",")
                    .lstrip("\n")
                )
            except UnicodeDecodeError as exception:
                message = f"Error decoding: {exception}"
                LOGGER.error(message)
                raise exception

            nvidia_smi_query_gpu_csv_dict_reader = csv.DictReader(
                io.StringIO(nvidia_smi_query_gpu_csv_decoded)
            )

            for gpu_info in nvidia_smi_query_gpu_csv_dict_reader:
                # Refactor key into B
                memory_total_in_mebibytes = gpu_info.pop("memory.total [MiB]")
                memory_size = MemorySize(memory_total_in_mebibytes)
                gpu_info["memory_total"] = memory_size.to_bytes()

                gpu_config.append(gpu_info)

        if platform.system() == "Darwin":
            # Check Metal gpu availability
            supported_metal_device = self._get_supported_metal_device()
            if supported_metal_device is not None:
                # Since Apple's SoC contains Metal,
                # we query the system itself for total memory
                system_profiler_hardware_data_type_command = (
                    "system_profiler SPHardwareDataType -json"
                )

                try:
                    system_profiler_hardware_data_type_output = subprocess.check_output(
                        system_profiler_hardware_data_type_command.split(" ")
                    )
                except subprocess.CalledProcessError as exception:
                    message = f"Error running {system_profiler_hardware_data_type_command}: {exception}"  # noqa
                    LOGGER.error(message)
                    raise exception

                try:
                    system_profiler_hardware_data_type_json = json.loads(
                        system_profiler_hardware_data_type_output
                    )
                except json.JSONDecodeError as exception:
                    message = f"Error decoding JSON: {exception}"  # noqa
                    LOGGER.error(message)
                    raise exception

                metal_device_json = system_profiler_hardware_data_type_json[
                    "SPHardwareDataType"
                ][supported_metal_device]

                gpu_info = {}
                gpu_info["name"] = metal_device_json.get("chip_type")

                # Refactor key into B
                physical_memory = metal_device_json.get("physical_memory")
                memory_size = MemorySize(physical_memory)
                gpu_info["memory_total"] = memory_size.to_bytes()

                gpu_config.append(gpu_info)

        # Raise an error if there is no valid gpu config
        if not gpu_config:
            message = "No valid gpu configuration"
            LOGGER.error(message)
            raise NotImplementedError(message)

        return gpu_config

    def get_windows_computer_service_product_values(self) -> Dict[str, str]:
        windows_computer_service_product_csv_command = (
            "cmd.exe /C wmic csproduct get Name, Vendor, Version, UUID /format:csv"
        )
        windows_computer_service_product_csv_output = subprocess.check_output(
            windows_computer_service_product_csv_command.split(" "),
            stderr=subprocess.DEVNULL,
        )
        windows_computer_service_product_csv_decoded = (
            windows_computer_service_product_csv_output.decode("utf-8")
            .replace("\r", "")
            .lstrip("\n")
        )
        windows_computer_service_product_dict = csv.DictReader(
            io.StringIO(windows_computer_service_product_csv_decoded)
        )
        csp_info = list(windows_computer_service_product_dict)[0]
        return {
            "windows_model_name": csp_info.get("Name", ""),
            "windows_model_vendor": csp_info.get("Vendor", ""),
            "windows_model_version": csp_info.get("Version", ""),
            "windows_model_uuid": csp_info.get("UUID", ""),
        }

    def get_windows_cpu_values(self) -> Dict[str, str]:
        windows_cpu_csv_command = (
            "cmd.exe /C wmic cpu get Name, MaxClockSpeed /format:csv"  # noqa
        )
        windows_cpu_csv_output = subprocess.check_output(
            windows_cpu_csv_command.split(" "),
            stderr=subprocess.DEVNULL,
        )
        windows_cpu_csv_decoded = (
            windows_cpu_csv_output.decode("utf-8").replace("\r", "").lstrip("\n")
        )
        windows_cpu_dict = csv.DictReader(io.StringIO(windows_cpu_csv_decoded))
        cpu_info = list(windows_cpu_dict)[0]
        return {
            "cpu_brand": cpu_info.get("Name", "").strip(),
            "cpu_max_clock_speed": cpu_info.get("MaxClockSpeed", ""),
        }

    def get_ubuntu_values(self) -> Dict[str, str]:
        get_machine_id_command = "cat /etc/machine-id"
        machine_id = subprocess.check_output(
            get_machine_id_command.split(" "),
            stderr=subprocess.DEVNULL,
        ).decode("utf-8")
        if machine_id:
            return {"linux_machine_id": machine_id}
        return {}

    def get_system_info(self):
        os_family = platform.system()
        system_info = {}
        if os_family == "Darwin":
            system_info = {**system_info, **self.get_darwin_system_profiler_values()}
            system_info["cpu_brand"] = (
                subprocess.check_output(["sysctl", "-n", "machdep.cpu.brand_string"])
                .strip()
                .decode("utf-8")
            )
            system_info["apple_mac_os_version"] = platform.mac_ver()[0]
        elif os_family == "Linux":
            # Support for Linux-based VMs in Windows
            if "WSL2" in platform.platform():
                system_info = {
                    **system_info,
                    **self.get_windows_computer_service_product_values(),
                    **self.get_windows_cpu_values(),
                }
            else:
                system_info = {**system_info, **self.get_ubuntu_values()}
        elif os_family == "Windows":
            system_info = {
                **system_info,
                **self.get_windows_computer_service_product_values(),
                **self.get_windows_cpu_values(),
            }

        system_info["name"] = platform.node()
        system_info["os_family"] = os_family
        system_info["os_release"] = platform.release()
        system_info["os_version"] = platform.version()
        system_info["platform"] = platform.platform()
        system_info["processor"] = platform.processor()
        system_info["machine"] = platform.machine()
        system_info["architecture"] = platform.architecture()[0]
        system_info["cpu_cores"] = str(platform.os.cpu_count())  # type: ignore exits
        system_info["gpu_config"] = self.get_gpu_config()
        return system_info

    def register(self):
        system_info = self.get_system_info()
        mutation = gql(
            """
            mutation registerHardware($input: RegisterHardwareInput!) {
                registerHardware(input: $input) {
                    ... on RegisterHardware {
                        fingerprint
                        rabbitmqPassword
                        rabbitmqHost
                        rabbitmqPort
                    }
                }
            }
        """
        )
        input = {"systemInfo": system_info}
        variables = {"input": input}
        result = self.spice.session.execute(mutation, variable_values=variables)
        fingerprint = result.get("registerHardware").get("fingerprint")
        rabbitmq_password = result.get("registerHardware").get("rabbitmqPassword")
        rabbitmq_host = result.get("registerHardware").get("rabbitmqHost")
        rabbitmq_port = result.get("registerHardware").get("rabbitmqPort")

        self.spice.host_config["fingerprint"] = fingerprint
        self.spice.host_config["rabbitmq_password"] = rabbitmq_password
        self.spice.host_config["rabbitmq_host"] = rabbitmq_host
        self.spice.host_config["rabbitmq_port"] = rabbitmq_port
        self.spice.full_config[self.spice.host] = self.spice.host_config
        update_config_file(new_config=self.spice.full_config)

        # then check in that the hardware, validate that it is saved correctly
        # and headers are
        self.spice.create_session()
        self.check_in_http(is_healthy=True)
        return result

    def update_hardware_system_info(self):
        """
        Updates hardware system information and returns the GraphQL response.

        Returns:
            dict: GraphQL response
        Raises:
            Exception: If no fingerprint is found or an error occurs
        """

        fingerprint = self.spice.host_config.get("fingerprint", None)
        if not fingerprint:
            message = "No fingerprint found. Please register: spice hardware register"
            raise Exception(message)

        system_info = self.get_system_info()
        new_state = {
            "systemInfo": system_info,
        }

        mutation = gql(
            """
            mutation updateHardwareSystemInfo($input: UpdateHardwareSystemInfoInput!) {
                updateHardwareSystemInfo(input: $input) {
                    ... on Hardware {
                        systemInfo
                    }
                }
            }
            """
        )

        input = new_state
        variables = {"input": input}
        try:
            result = self.spice.session.execute(mutation, variable_values=variables)
        except client_exceptions.ClientConnectorError as exception:
            message = " [*] Failed to update hardware system info! "
            LOGGER.error(message)
            raise exception

        message = " [*] Updated hardware system info successfully! "
        LOGGER.info(message)

        return result

    def check_in_http(
        self,
        is_healthy: bool = True,
        is_quarantined: bool = False,
        is_available: bool = False,
        is_online: bool = True,
    ):
        fingerprint = self.spice.host_config.get("fingerprint", None)
        if not fingerprint:
            message = "No fingerprint found. Please register: spice hardware register"
            raise Exception(message)

        new_state = {
            "isHealthy": is_healthy,
            "isQuarantined": is_quarantined,
            "isAvailable": is_available,
            "isOnline": is_online,
        }

        mutation = gql(
            """
            mutation checkIn($input: CheckInInput!) {
                checkIn(input: $input) {
                    ... on Hardware {
                        createdAt
                        updatedAt
                        lastCheckIn
                    }
                }
            }
        """  # noqa
        )
        input = new_state
        variables = {"input": input}
        try:
            result = self.spice.session.execute(mutation, variable_values=variables)
            previous_state = self.previous_state
            self.previous_state = new_state.copy()

            message = " [*] Checked in successfully: "
            for key, value in new_state.items():
                if value != previous_state.get(key, None):
                    if value is True:
                        message = (
                            message
                            + click.style(f"{key}: ")
                            + click.style("💤")
                            + click.style(" ==> ✅ ", fg="green")
                        )
                    else:
                        message = (
                            message
                            + click.style(f"{key}: ")
                            + click.style("✅")
                            + click.style(" ==> 💤 ", fg="yellow")
                        )
            LOGGER.info(message)

            return result
        except client_exceptions.ClientConnectorError:
            config = read_config_file(filepath=SPICE_TRAINING_FILEPATH)
            if config.get("status") in self.spice.worker.ACTIVE_STATUSES:
                return None



# tests/test_memory_size.py

import pytest
from spice_agent.utils.memory_size import MemorySize


def test_invalid_value_initialization():
    # Unit should never be None in a value-based initialization
    with pytest.raises(ValueError):
        MemorySize(512, None)

    # An unsupported unit type should never be passed int
    with pytest.raises(ValueError):
        MemorySize(512, "invalid_unit")


def test_invalid_string_initialization():
    # There should be a space between the value and the unit
    with pytest.raises(ValueError):
        MemorySize("1.5TB")

    # No numeric value
    with pytest.raises(ValueError):
        MemorySize("invalid_unit MB")

    with pytest.raises(ValueError):
        MemorySize("3.14 xyz")  # Invalid unit


def test_to_bytes_conversion():
    # Check same unit
    size = MemorySize(512, "B")
    assert size.to_bytes() == 512

    size = MemorySize("512 B")
    assert size.to_bytes() == 512

    # Check going up in scale
    size = MemorySize(1024, "b")
    assert size.to_bytes() == 128

    size = MemorySize("1024 b")
    assert size.to_bytes() == 128

    # Check going down in scale
    size = MemorySize(1.03, "GB")
    assert size.to_bytes() == 1.03 * 1000**3

    size = MemorySize("1.03 GB")
    assert size.to_bytes() == 1.03 * 1000**3

    # Check going from binary to decimal
    size = MemorySize(2.123, "PiB")
    assert size.to_bytes() == 2.123 * 2**50

    size = MemorySize("2.123 PiB")
    assert size.to_bytes() == 2.123 * 2**50


def test_convert_to():
    size = MemorySize(512, "MiB")

    # Check same unit
    size.convert_to("MiB")
    assert str(size) == "512.000 MiB"

    # Convert down
    size.convert_to("KiB")
    assert str(size) == "524288.000 KiB"

    # Convert up
    size.convert_to("GiB")
    assert str(size) == "0.500 GiB"

    # Convert from binary to decimal
    size.convert_to("B")
    assert str(size) == "536870912.000 B"

    # Test invalid unit
    with pytest.raises(ValueError):
        size.convert_to("invalid")


def test_to_human_readable():
    size = MemorySize(123123, "MB")
    value, unit = size.get_human_readable()
    assert f"{value:.3f} {unit}" == "123.123 GB"

    size = MemorySize(314159, "MiB")
    value, unit = size.get_human_readable()
    assert f"{value:.3f} {unit}" == "329.420 GB"


def test_comparison_equal():
    size1 = MemorySize(1000, "KB")
    size2 = MemorySize(1, "MB")
    assert size1 == size2

    size1 = MemorySize(1024, "KiB")
    size2 = MemorySize(1, "MiB")
    assert size1 == size2


def test_comparison_not_equal():
    size1 = MemorySize(1024, "B")
    size2 = MemorySize(1, "KB")
    assert size1 != size2


def test_comparison_type_error():
    size = MemorySize(512, "B")
    with pytest.raises(TypeError):
        result = size == "512 B"
        assert result is False



from importlib.metadata import version
import logging
import os
import platform
import sys
from typing import Tuple

from outdated import check_outdated

from spice_agent.daemons.actions import Daemons

LOGGER = logging.getLogger(__name__)


def get_current_version():
    return version("spice_agent")


def get_is_outdated_and_get_latest_version(current_version: str) -> Tuple[bool, str]:
    if not current_version:
        current_version = get_current_version()
    is_outdated, latest_version = check_outdated("spice_agent", current_version)
    return is_outdated, latest_version


def update_if_outdated():
    current_version = get_current_version()
    is_outdated, latest_version = get_is_outdated_and_get_latest_version(
        current_version=current_version
    )
    os_family = platform.system()
    if is_outdated:
        LOGGER.warn(
            f"spice_agent version is at: {get_current_version()}. Version {latest_version} is available."  # noqa
        )
        # currently only supporting macOS and WSL2
        if os_family == "Darwin":
            LOGGER.warn("Attempting to update and restart the spice daemon.")
            os.system("pip install --upgrade spice-agent")
            daemons = Daemons(spice=None)
            daemons.uninstall()
            daemons.install()
            sys.exit()
        elif os_family == "Linux" and "WSL2" in platform.platform():
            LOGGER.warn("Attempting to update and restart the spice daemon.")
            os.system("pip install --upgrade spice-agent")
            daemons = Daemons(spice=None)
            daemons.uninstall()
            daemons.install()
            sys.exit()
        else:
            LOGGER.warn(
                """Please update via:
pip install --upgrade spice-agent"""
            )



import json
import os
from pathlib import Path
import shutil
from typing import Dict

HOME_DIRECTORY = Path.home()
SPICE_INFERENCE_DIRECTORY = Path(HOME_DIRECTORY / ".cache" / "spice" / "inference")
SPICE_HOSTS_FILEPATH = Path(HOME_DIRECTORY / ".config" / "spice" / "hosts.json")
SPICE_TRAINING_FILEPATH = Path(HOME_DIRECTORY / ".config" / "spice" / "training.json")
SPICE_ROUND_VERIFICATION_FILEPATH = Path(
    HOME_DIRECTORY / ".config" / "spice" / "training-verification.json"
)
SPICE_MODEL_CACHE_FILEPATH = Path(HOME_DIRECTORY / ".cache" / "spice" / "models")
HF_HUB_DIRECTORY = Path(HOME_DIRECTORY / ".cache" / "huggingface" / "hub")


def create_directory(filepath: Path = SPICE_MODEL_CACHE_FILEPATH):
    filepath.mkdir(parents=True, exist_ok=True)


def create_config_file(filepath: Path = SPICE_HOSTS_FILEPATH):
    filepath.parent.mkdir(parents=True, exist_ok=True)
    filepath.touch()
    with filepath.open("w") as json_file:
        json.dump({}, json_file)


def update_config_file(filepath: Path = SPICE_HOSTS_FILEPATH, new_config: Dict = {}):
    existing_config = read_config_file(filepath=filepath)
    merged_config = {**existing_config, **new_config}
    with filepath.open("w") as json_file:
        json.dump(merged_config, json_file, indent=2)


def read_config_file(filepath: Path = SPICE_HOSTS_FILEPATH) -> Dict:
    if not filepath.exists():
        create_config_file(filepath=filepath)
    return json.loads(filepath.read_text())


def copy_directory(src_dir, dst_dir, ignore_files=["config.json", "pytorch_model.bin"]):
    """
    Copies current training_round_model_repo config files (e.g. tokenizer, vocab)
    into model cache for upload to s3 and hf
    """
    for file in os.listdir(src_dir):
        if file in ignore_files:
            continue

        src_path = Path(src_dir / file)
        dst_path = Path(dst_dir / file)
        shutil.copy2(src_path, dst_path)



import click

from ..utils.printer import print_result


@click.group()
@click.pass_context
def cli(context):
    """Worker"""
    pass


@cli.command("start")
@click.pass_context
def start(context):
    """Run as a worker that picks up tasks from the spice.cloud queue."""
    spice = context.obj.get("SPICE")
    message = spice.worker.start()
    print_result(message=message, context=context)



import hashlib
from pathlib import Path
import sys
import threading
from typing import Dict, Optional

import boto3
from boto3.s3.transfer import TransferConfig
import botocore.exceptions
from gql import gql

from spice_agent.graphql.sdk import create_requests_session


class ProgressPercentage(object):
    def __init__(self, filepath: Path) -> None:
        self._filename = filepath.name
        self._size = float(filepath.stat().st_size)
        self._seen_so_far = 0
        self._lock = threading.Lock()

    def __call__(self, bytes_amount):
        # To simplify, assume this is hooked up to a single filename
        with self._lock:
            self._seen_so_far += bytes_amount
            percentage = (self._seen_so_far / self._size) * 100
            sys.stdout.write(
                "\r%s  %s / %s  (%.2f%%)"
                % (self._filename, self._seen_so_far, self._size, percentage)
            )
            sys.stdout.flush()


class Uploader:
    def __init__(self, spice) -> None:
        self.spice = spice

    def _get_access(self):
        query = gql(
            """
            query getAgentUploaderAccess {
                getAgentUploaderAccess {
                    accessKeyId
                    secretAccessKey
                }
            }
        """  # noqa
        )
        result = self.spice.session.execute(query)
        return {
            "access_key_id": result["getAgentUploaderAccess"]["accessKeyId"],
            "secret_access_key": result["getAgentUploaderAccess"]["secretAccessKey"],
        }

    def _create_file(
        self, file_name: str, file_size: int, file_checksum: str, location: str
    ):
        mutation = gql(
            """
            mutation createFile($input: CreateFileInput!) {
                createFile(input: $input) {
                    ... on File {
                        id
                    }
                }
            }
        """  # noqa
        )
        input = {
            "fileName": file_name,
            "fileSize": file_size,
            "fileChecksum": file_checksum,
            "location": location,
        }
        variables = {"input": input}
        result = self.spice.session.execute(mutation, variable_values=variables)
        return result.get("createFile").get("id")

    def _update_file_status(
        self,
        file_id: str,
        is_uploading: Optional[bool] = None,
        is_complete: Optional[bool] = None,
    ):
        mutation = gql(
            """
            mutation updateFileStatus($input: UpdateFileStatusInput!) {
                updateFileStatus(input: $input) {
                    ... on File {
                        id
                    }
                }
            }
        """  # noqa
        )
        input: Dict[str, str | bool] = {
            "fileId": file_id,
        }
        if is_uploading is not None:
            input["isUploading"] = is_uploading
        if is_complete is not None:
            input["isComplete"] = is_complete

        variables = {"input": input}
        result = self.spice.session.execute(mutation, variable_values=variables)
        return result

    def upload_file_direct(
        self,
        bucket_name: str,
        key: str,
        filepath: Path,
        file_id: Optional[str],
        overwrite: bool = False,
    ):
        """
        bucket: name of S3 bucket ie spice-models
        key: dirs and file inside bucket /some/dir/file.json
        """
        if filepath.exists() is False:
            raise Exception(f"File {filepath} does not exist.")

        file_size = filepath.stat().st_size
        if file_size == 0:
            raise Exception(f"Model path {filepath} is empty.")

        file_checksum = hashlib.md5(filepath.read_bytes()).hexdigest()

        if not file_id:
            file_id = self._create_file(
                file_name=filepath.name,
                file_size=filepath.stat().st_size,
                file_checksum=file_checksum,
                location=f"s3://{key}",
            )

        if not file_id:
            raise Exception("No file_id found or provided.")

        s3_access_values = self._get_access()
        s3_client = boto3.client(
            "s3",
            aws_access_key_id=s3_access_values["access_key_id"],
            aws_secret_access_key=s3_access_values["secret_access_key"],
        )

        self._update_file_status(file_id=file_id, is_uploading=True, is_complete=False)

        file_already_uploaded = False
        try:
            head_object = s3_client.head_object(Bucket=bucket_name, Key=key)

            # TODO: add a check here that the checksum is the same as well
            if (
                head_object["ResponseMetadata"]["HTTPHeaders"]["content-length"]
                == file_size
            ):
                file_already_uploaded = True
                print("File already uploaded.")
            else:
                print("File in bucket has a file size mismatch. Reuploading.")
        except botocore.exceptions.ClientError as exception:
            if exception.response["Error"]["Code"] in ["404", "403"]:
                # The key does not exist and we should upload the file
                pass
            # elif exception.response["Error"]["Code"] == "403":
            #     # Unauthorized, including invalid bucket
            #     # raise exception
            #     raise Exception("Unauthorized to view bucket.")
            else:
                print("Unhandled exception.")
                raise exception

        if (file_already_uploaded is False) or (
            file_already_uploaded and overwrite is True
        ):
            # TODO: configure these values based on the available system properties
            # if a file is bigger than multipart_threshold, then do multipart upload
            # multipart_threshold = 1024 * 100
            # multipart_chunksize = 1024 * 100
            max_concurrency = 16
            config = TransferConfig(
                # multipart_threshold=multipart_threshold,
                max_concurrency=max_concurrency,
                # multipart_chunksize=multipart_chunksize,  # 25MB
                use_threads=True,
            )
            s3_client.upload_file(
                filepath,
                bucket_name,
                key,
                Config=config,
                Callback=ProgressPercentage(filepath=filepath),
            )

        print("")  # print out a newline
        self._update_file_status(file_id=file_id, is_uploading=False, is_complete=True)

    def upload_file_via_api(self, path: Path):
        operations = """{ "query": "mutation uploadFile($input: UploadFileInput!) { uploadFile(input: $input) { ... on File { id } } }", "variables": { "input": { "fileObject": null } } }"""  # noqa
        body = {
            "operations": ("", operations),
            "map": ("", '{"fileObject": ["variables.input.fileObject"]}'),
            "fileObject": (path.name, open(path, "rb")),
        }

        session = create_requests_session(self.spice.host_config)
        transport = self.spice.host_config.get("transport")
        url = f"{transport}://{self.spice.host}/"
        response = session.post(url, files=body)
        return response



import hashlib
import json
import logging
import os
import threading
from typing import Dict, Optional

from aiohttp import client_exceptions
from datasets import load_dataset
import evaluate
from gql import gql
from gql.transport.exceptions import TransportQueryError
import numpy as np
import requests
import torch  # noqa
from torchvision.transforms import Compose, Normalize, RandomResizedCrop, ToTensor
from transformers import (  # noqa
    AutoImageProcessor,
    AutoModelForImageClassification,
    AutoModelForSequenceClassification,
    AutoTokenizer,
    Trainer,
    TrainingArguments,
)
from transformers.data import DefaultDataCollator

from spice_agent.utils.config import (
    HF_HUB_DIRECTORY,
    SPICE_MODEL_CACHE_FILEPATH,
    SPICE_ROUND_VERIFICATION_FILEPATH,
    SPICE_TRAINING_FILEPATH,
    copy_directory,
    create_directory,
    read_config_file,
    update_config_file,
)

os.environ["PYTORCH_ENABLE_MPS_FALLBACK"] = "1"

import torch  # noqa
import transformers  # noqa

LOGGER = logging.getLogger(__name__)


MODEL_BUCKET_NAME = "spice-models"

TOKENIZER_MAX_LENGTH = 32


class ThreadedStatusDetailsCallbackDecorator:
    """
    A class decorator that enables threaded functionality for StatusDetailsCallback
    """

    def __init__(self, is_threaded: Optional[bool] = False):
        self.is_threaded = is_threaded
        self.current_thread = None

    def __call__(
        self,
        cls,
        functions_to_decorate=["on_train_begin", "on_step_end", "on_train_end"],
    ):
        if not self.is_threaded:
            return cls

        # Wrap functions with defined decorators
        # Example: on_step_end gets wrapped with on_step_end_decorator
        decorated_methods = []
        for name in functions_to_decorate:
            value = cls.__dict__[name]
            if callable(value):
                decorated_method = getattr(self, f"{name}_decorator", None)
                if decorated_method:
                    decorated_method = decorated_method(value)
                    setattr(cls, name, decorated_method)
                    decorated_methods.append(name)

        # Set wrapped class' attributes
        setattr(cls, "decorated_methods", decorated_methods)
        return cls

    def __del__(self):
        self._clean_current_thread()

    def on_train_begin_decorator(self, function):
        def wrapper(*args, **kwargs):
            try:
                if not self.current_thread:
                    self.current_thread = self._get_threaded_function(
                        function, *args, **kwargs
                    )

                    self.current_thread.start()
            except KeyboardInterrupt as exception:
                self._clean_current_thread()
                raise exception

        return wrapper

    def on_step_end_decorator(self, function):
        def wrapper(*args, **kwargs):
            try:
                # If the current thread is no longer alive, we join it with the main
                # thread and reset current thread
                if self.current_thread and not self.current_thread.is_alive():
                    self.current_thread.join()
                    del self.current_thread
                    self.current_thread = None

                # We set the current thread and start it's activity
                if not self.current_thread:
                    self.current_thread = self._get_threaded_function(
                        function, *args, **kwargs
                    )
                    self.current_thread.start()
            except KeyboardInterrupt as exception:
                self._clean_current_thread()
                raise exception

        return wrapper

    def on_train_end_decorator(self, function):
        def wrapper(*args, **kwargs):
            try:
                # Complete final on_step_end update
                if self.current_thread and self.current_thread.is_alive():
                    self.current_thread.join()
                    del self.current_thread
                    self.current_thread = None

                # We execute on_train_end on the main thread since it is the final call
                # in StatusDetailsCallback and we have to wait for it's completion.
                function(*args, **kwargs)
            except KeyboardInterrupt as exception:
                self._clean_current_thread()
                raise exception

        return wrapper

    def _get_threaded_function(self, function, *args, **kwargs) -> threading.Thread:
        def target_function():
            function(*args, **kwargs)

        thread = threading.Thread(target=target_function)
        return thread

    def _clean_current_thread(self):
        if self.current_thread:
            if self.current_thread.is_alive():
                self.current_thread.join()
            del self.current_thread


@ThreadedStatusDetailsCallbackDecorator(is_threaded=True)
class StatusDetailsCallback(transformers.TrainerCallback):
    """
    A [`TrainerCallback`] that sends the progress of training or evaluation
    to the spice backend.
    """

    def __init__(
        self,
        training,
        status: str,
        training_round_step_id: Optional[str] = None,
        training_round_id: Optional[str] = None,
    ):
        self.training = training
        self.status = status
        self.training_round_step_id = training_round_step_id
        self.training_round_id = training_round_id

    def on_train_begin(self, args, state, control, **kwargs):
        if state.is_local_process_zero:
            if self.training_round_step_id:
                self.training._update_training_round_step(
                    training_round_step_id=self.training_round_step_id,
                    status=self.status,
                    status_details={"progress": 0},
                )
            if self.training_round_id:
                self.training._update_training_round(
                    training_round_id=self.training_round_id,
                    status=self.status,
                    status_details={"progress": 0},
                )
        self.current_step = 0

    def on_step_end(self, args, state, control, **kwargs):
        if state.is_local_process_zero:
            progress = round((state.global_step / state.max_steps) * 100)
            if self.training_round_step_id:
                self.training._update_training_round_step(
                    training_round_step_id=self.training_round_step_id,
                    status=self.status,
                    status_details={"progress": progress},
                )
            if self.training_round_id:
                self.training._update_training_round(
                    training_round_id=self.training_round_id,
                    status=self.status,
                    status_details={"progress": progress},
                )
            self.current_step = state.global_step

    def on_train_end(self, args, state, control, **kwargs):
        if state.is_local_process_zero:
            if self.training_round_step_id:
                self.training._update_training_round_step(
                    training_round_step_id=self.training_round_step_id,
                    status=self.status,
                    status_details={"progress": 100},
                )
            if self.training_round_id:
                self.training._update_training_round(
                    training_round_id=self.training_round_id,
                    status=self.status,
                    status_details={"progress": 100},
                )


class Training:
    def __init__(self, spice) -> None:
        self.spice = spice
        self.device = self.spice.get_device()

        # logging.basicConfig(level=logging.INFO)
        if self.spice.DEBUG:
            transformers.logging.set_verbosity_debug()
            logging.getLogger("transformers.modeling_utils").setLevel(logging.ERROR)
            logging.getLogger("pika").setLevel(logging.ERROR)

    def _get_training_round_step(self, training_round_step_id):
        query = gql(
            """
            query trainingRoundStep($id: GlobalID!) {
                trainingRoundStep(id: $id) {
                    id
                    status
                }
            }
            """  # noqa
        )
        variables = {
            "id": training_round_step_id,
        }

        return self.spice.session.execute(query, variable_values=variables)

    def _get_training_round(self, training_round_id):
        query = gql(
            """
            query trainingRound($id: GlobalID!) {
                trainingRound(id: $id) {
                    id
                    status
                }
            }
            """  # noqa
        )
        variables = {
            "id": training_round_id,
        }

        return self.spice.session.execute(query, variable_values=variables)

    def _update_training_round(
        self,
        training_round_id: str,
        status: str,
        status_details: Optional[dict] = None,
        metrics: Optional[dict] = None,
    ):
        mutation = gql(
            """
            mutation updateTrainingRound($input: UpdateTrainingRoundInput!) {
                updateTrainingRound(input: $input) {
                    ... on TrainingRound {
                        id
                        status
                        statusDetails
                        roundNumber
                        trainingJob {
                            id
                            baseModel {
                                repoId
                                repoRevision
                            }
                            baseDatasetRepoId
                            baseDatasetRepoRevision
                            hfModel {
                                repoId
                                repoRevision
                            }
                        }
                    }
                }
            }
            """  # noqa
        )
        input: Dict[str, str | float] = {"trainingRoundId": training_round_id}
        if status is not None:
            input["status"] = status
        if status_details:
            input["statusDetails"] = json.dumps(status_details)
        if metrics:
            input["metrics"] = json.dumps(metrics)

        variables = {"input": input}
        try:
            result = self.spice.session.execute(mutation, variable_values=variables)
        except TransportQueryError as exception:
            if exception.errors:
                for error in exception.errors:
                    if error.get("message", None) == "Round not found.":
                        LOGGER.error("Round not found.")
                        return None
            else:
                raise exception

        update_config_file(
            filepath=SPICE_ROUND_VERIFICATION_FILEPATH,
            new_config=result["updateTrainingRound"],
        )
        return result

    def _update_training_round_step(
        self,
        training_round_step_id: str,
        status: str,
        status_details: Optional[dict] = None,
        file_id: Optional[str] = None,
        metrics: Optional[dict] = None,
    ):
        mutation = gql(
            """
            mutation updateTrainingRoundStep($input: UpdateTrainingRoundStepInput!) {
                updateTrainingRoundStep(input: $input) {
                    ... on TrainingRoundStep {
                        id
                        status
                        hfModelRepoId
                        hfModelRepoRevision
                        hfDatasetRepoId
                        hfDatasetRepoRevision
                        datasetStartingRow
                        datasetEndingRow
                        trainingEpochs
                        trainingBatchSize
                        trainingRound {
                            id
                            roundNumber
                            trainingJob {
                                id
                                baseModel {
                                    repoId
                                    repoRevision
                                }
                            }
                        }
                        statusDetails
                    }
                }
            }
        """  # noqa
        )
        input: Dict[str, str | float] = {"trainingRoundStepId": training_round_step_id}
        if status is not None:
            input["status"] = status
        if status_details:
            input["statusDetails"] = json.dumps(status_details)
        if file_id is not None:
            input["fileId"] = file_id
        if metrics:
            input["metrics"] = json.dumps(metrics)

        variables = {"input": input}

        try:
            result = self.spice.session.execute(mutation, variable_values=variables)
        except TransportQueryError as exception:
            if exception.errors:
                for error in exception.errors:
                    if error.get("message", None) == "Round step not found.":
                        LOGGER.error("Round step not found.")
                        return None
            else:
                raise exception
        except client_exceptions.ClientOSError:
            # the backend can be down or deploying this step, do not break the app
            if status in self.spice.worker.ACTIVE_STATUSES:
                return None

        update_config_file(
            filepath=SPICE_TRAINING_FILEPATH,
            new_config=result["updateTrainingRoundStep"],
        )
        return result

    def _get_agent_round_presigned_urls(self, training_round_id: str):
        # todo: remove training_round_number
        query = gql(
            """
            query getAgentRoundPresignedUrls($trainingRoundId: GlobalID!) {
                getAgentRoundPresignedUrls(trainingRoundId: $trainingRoundId) {
                    roundModel
                    config
                }
            }
            """  # noqa
        )
        variables = {
            "trainingRoundId": training_round_id,
        }

        return self.spice.session.execute(query, variable_values=variables)

    def upload_models(self):
        config = read_config_file(filepath=SPICE_TRAINING_FILEPATH)
        training_round_step_id = config["id"]
        if not self.check_training_round_step_exists(training_round_step_id):
            return None

        hf_model_repo_id = config["hfModelRepoId"]
        training_round_id = config["trainingRound"]["id"]
        training_round_number = config["trainingRound"]["roundNumber"]
        training_job_id = config["trainingRound"]["trainingJob"]["id"]
        training_round_directory = f"{training_round_id}/steps/{training_round_step_id}"

        # round_id_step_id will be in form: [uuid]/steps/[uuid]
        # round_id_step_id is used for finding the step model in the cache
        # and as the key in S3
        bucket_dir = f"{training_job_id}/rounds/{training_round_number}/steps/{training_round_step_id}/"  # noqa

        # model_cache_for_training_round contains all the step models
        # for this particular training round
        model_cache_for_training_round = SPICE_MODEL_CACHE_FILEPATH.joinpath(
            training_round_directory
        )

        self._update_training_round_step(
            training_round_step_id=training_round_step_id,
            status="UPLOADING",
        )

        # hf_model_directory could refer to different snapshots of the model
        hf_model_repo_id_formatted = hf_model_repo_id.replace("/", "--")
        snapshot_directory = f"models--{hf_model_repo_id_formatted}/snapshots"
        hf_model_directory = HF_HUB_DIRECTORY.joinpath(snapshot_directory)
        recent_snapshot_directory = max(
            hf_model_directory.iterdir(), key=lambda d: d.stat().st_mtime
        )
        recent_hf_model_directory = hf_model_directory.joinpath(
            recent_snapshot_directory
        )
        copy_directory(recent_hf_model_directory, model_cache_for_training_round)

        for file in model_cache_for_training_round.iterdir():
            if file.is_file():
                bucket_key = bucket_dir + file.name

                # create the "file" and attach it to the training round step
                file_checksum = hashlib.md5(file.read_bytes()).hexdigest()
                file_id = self.spice.uploader._create_file(
                    file_name=file.name,
                    file_size=file.stat().st_size,
                    file_checksum=file_checksum,
                    location=f"s3://{bucket_key}",
                )
                self.spice.uploader.upload_file_direct(
                    bucket_name=MODEL_BUCKET_NAME,
                    key=bucket_key,
                    filepath=file,
                    file_id=file_id,
                )

        self._update_training_round_step(
            training_round_step_id=training_round_step_id, status="COMPLETE"
        )

    def _load_dataset(self, config, task):
        # TODO: create one central training config file
        if task == "train" or task == "test":
            hf_dataset_repo_id = config["hfDatasetRepoId"]
            hf_dataset_repo_revision = config["hfDatasetRepoRevision"]
        elif task == "verify":
            hf_dataset_repo_id = config["trainingJob"]["baseDatasetRepoId"]
            hf_dataset_repo_revision = config["trainingJob"]["baseDatasetRepoRevision"]
        else:
            error_message = f"_load_dataset with task={task} does not exist!"
            LOGGER.error(error_message)
            raise ValueError(error_message)

        if task == "train":
            dataset_starting_row = config["datasetStartingRow"]
            dataset_ending_row = config["datasetEndingRow"]
            split = f"train[{dataset_starting_row}:{dataset_ending_row}]"
        else:
            split = "test"  # both test_model and verify_model uses test split

        return load_dataset(
            path=hf_dataset_repo_id, revision=hf_dataset_repo_revision, split=split
        )

    def _tokenize_dataset(self, dataset, config, task):
        # TODO: create one central training config file

        # get tokenizer
        print("Loading tokenizer...")
        if task == "train" or task == "test":
            hf_model_repo_id = config["hfModelRepoId"]
            hf_model_repo_revision = config["hfModelRepoRevision"]

            tokenizer = AutoTokenizer.from_pretrained(
                pretrained_model_name_or_path=hf_model_repo_id,
                revision=hf_model_repo_revision,
            )
        elif task == "verify":
            training_round_number = config["roundNumber"]
            hf_model_repo_id = config["trainingJob"]["hfModel"]["repoId"]
            base_model_repo_id = config["trainingJob"]["baseModel"]["repoId"]
            base_model_repo_revision = config["trainingJob"]["baseModel"][
                "repoRevision"
            ]

            tokenizer = AutoTokenizer.from_pretrained(
                pretrained_model_name_or_path=hf_model_repo_id
                if hf_model_repo_id and training_round_number > 1
                else base_model_repo_id,
                revision="main"  # uses main tokenizer
                if hf_model_repo_id and training_round_number > 1
                else base_model_repo_revision,
            )
        else:
            error_message = f"_tokenize_dataset with task={task} does not exist!"
            LOGGER.error(error_message)
            raise ValueError(error_message)

        # create a tokenize function that will tokenize the dataset
        def tokenize_function(examples):
            return tokenizer(
                examples["text"],
                padding="max_length",
                truncation=True,
                max_length=TOKENIZER_MAX_LENGTH,
            )

        print("Tokenizing dataset...")
        tokenized_dataset = dataset.map(tokenize_function, batched=True)

        # Remove the text column because the model does not accept raw text as an input
        tokenized_dataset = tokenized_dataset.remove_columns(["text"])

        # Rename the label column to labels because
        # the model expects the argument to be named labels
        tokenized_dataset = tokenized_dataset.rename_column("label", "labels")

        # Set the format of the dataset to return PyTorch tensors instead of lists
        tokenized_dataset.set_format("torch")

        return tokenizer, tokenized_dataset

    def _process_dataset(self, dataset, config, task):
        # get image processor
        LOGGER.info("Loading image processor...")
        if task in ["train", "test"]:
            hf_model_repo_id = config["hfModelRepoId"]
            hf_model_repo_revision = config["hfModelRepoRevision"]

            image_processor = AutoImageProcessor.from_pretrained(
                pretrained_model_name_or_path=hf_model_repo_id,
                revision=hf_model_repo_revision,
                trust_remote_code=True,
            )
        elif task == "verify":
            training_round_number = config["roundNumber"]
            hf_model_repo_id = config["trainingJob"]["hfModel"]["repoId"]
            base_model_repo_id = config["trainingJob"]["baseModel"]["repoId"]
            base_model_repo_revision = config["trainingJob"]["baseModel"][
                "repoRevision"
            ]

            pretrained_model_name_or_path = base_model_repo_id
            revision = "main"
            if hf_model_repo_id and training_round_number > 1:
                pretrained_model_name_or_path = base_model_repo_id
                revision = base_model_repo_revision

            image_processor = AutoImageProcessor.from_pretrained(
                pretrained_model_name_or_path=pretrained_model_name_or_path,
                revision=revision,
                trust_remote_code=True,
            )
        else:
            error_message = f"_process_dataset with task={task} does not exist!"
            LOGGER.error(error_message)
            raise ValueError(error_message)

        normalize = Normalize(
            mean=image_processor.image_mean, std=image_processor.image_std
        )

        size = image_processor.size.get("shortest_edge", None)
        if (
            not size
            and image_processor.size.get("height", None)
            and image_processor.size.get("width", None)
        ):
            size = (image_processor.size["height"], image_processor.size["width"])
        else:
            message = "size was not set. image_processor did not have keys \
                shortest_edge, or height and width."
            LOGGER.error(message)
            raise Exception(message)

        _transforms = Compose([RandomResizedCrop(size), ToTensor(), normalize])

        def transforms(examples):
            examples["tensor"] = [_transforms(img) for img in examples["image"]]
            del examples["image"]
            return examples

        print("Processing dataset...")
        dataset = dataset.with_transform(transforms)
        return image_processor, dataset

    def _load_model(self, config, task):
        # Setup model cache for training/testing
        model_cache_for_training_round_step = None
        base_model_repo_id = None
        if task == "train" or task == "test":
            training_round_id = config["trainingRound"]["id"]
            training_round_step_id = config["id"]
            base_model_repo_id = config["trainingRound"]["trainingJob"]["baseModel"][
                "repoId"
            ]
            training_round_directory = (
                f"{training_round_id}/steps/{training_round_step_id}"
            )
            model_cache_for_training_round_step = SPICE_MODEL_CACHE_FILEPATH.joinpath(
                training_round_directory
            )

        # Select pretrained model
        if task == "train":
            pretrained_model_name_or_path = config["hfModelRepoId"]
        elif task == "test":
            if not model_cache_for_training_round_step:
                error_message = f"_get_trainer with task={task} requires model_cache_for_training_round_step!"  # noqa
                LOGGER.error(error_message)
                raise ValueError(error_message)
            pretrained_model_name_or_path = model_cache_for_training_round_step
        elif task == "verify":
            training_round_id = config["id"]
            base_model_repo_id = config["trainingJob"]["baseModelRepoId"]
            spice_model_round_cache = SPICE_MODEL_CACHE_FILEPATH.joinpath(
                f"{training_round_id}/"
            )
            pretrained_model_name_or_path = spice_model_round_cache
        else:
            error_message = f"_load_model with task={task} does not exist!"
            LOGGER.error(error_message)
            raise ValueError(error_message)

        # Load your model with the number of expected labels:
        LOGGER.info(f"Loading {task} model...")
        if base_model_repo_id == "spicecloud/spice-cnn-base":
            model = AutoModelForImageClassification.from_pretrained(
                pretrained_model_name_or_path,
                num_labels=10,
                trust_remote_code=True,
            )
        elif base_model_repo_id == "bert-base-uncased":
            model = AutoModelForSequenceClassification.from_pretrained(
                pretrained_model_name_or_path, num_labels=5
            )
        else:
            error_message = (
                f"_load_model has unknown base model: {base_model_repo_id} type"
            )
            LOGGER.error(error_message)
            raise ValueError(error_message)

        return model

    def _get_training_arguments(self, output_dir, config, task):
        training_arguments = TrainingArguments(
            output_dir=str(output_dir),
            num_train_epochs=1,
            use_mps_device=torch.backends.mps.is_available(),  # type: ignore
        )

        if task == "train":
            # set to "no", set to "epoch" for eval + train
            training_arguments.evaluation_strategy = "no"
            training_arguments.per_device_train_batch_size = config["trainingBatchSize"]
        elif task == "test":
            training_arguments.evaluation_strategy = "epoch"
            training_arguments.per_device_eval_batch_size = config["trainingBatchSize"]
        elif task == "verify":
            training_arguments.evaluation_strategy = "epoch"
            training_arguments.per_device_eval_batch_size = 32
            training_arguments.save_strategy = "no"
        else:
            error_message = f"_get_training_arguments with task={task} does not exist!"
            LOGGER.error(error_message)
            raise ValueError(error_message)

        return training_arguments

    def train_model(self):
        config = read_config_file(filepath=SPICE_TRAINING_FILEPATH)
        training_round_step_id = config["id"]
        if not self.check_training_round_step_exists(training_round_step_id):
            return None

        training_round_id = config["trainingRound"]["id"]
        base_model_repo_id = config["trainingRound"]["trainingJob"]["baseModel"][
            "repoId"
        ]

        # create the folder for the new training round
        # where the step model will be saved
        training_round_directory = f"{training_round_id}/steps/{training_round_step_id}"
        model_cache_for_training_round = SPICE_MODEL_CACHE_FILEPATH.joinpath(
            training_round_directory
        )
        create_directory(filepath=model_cache_for_training_round)

        # get dataset
        print("Loading dataset...")
        self._update_training_round_step(
            training_round_step_id=training_round_step_id, status="DOWNLOADING_DATASET"
        )
        train_dataset = self._load_dataset(config, "train")

        # tokenize dataset
        tokenizer = None
        if base_model_repo_id == "spicecloud/spice-cnn-base":
            tokenizer, tokenized_dataset = self._process_dataset(
                train_dataset, config, "train"
            )
        elif base_model_repo_id == "bert-base-uncased":
            tokenizer, tokenized_dataset = self._tokenize_dataset(
                train_dataset, config, "train"
            )
        else:
            error_message = (
                f"train_model has unknown base model: {base_model_repo_id} type"
            )
            LOGGER.error(error_message)
            raise ValueError(error_message)

        # load your model with the number of expected labels:
        model = self._load_model(config, "train")

        training_args = self._get_training_arguments(
            model_cache_for_training_round, config, "train"
        )
        metric = evaluate.load("accuracy")

        def compute_metrics(eval_pred):
            logits, labels = eval_pred
            predictions = np.argmax(logits, axis=-1)
            return metric.compute(predictions=predictions, references=labels)

        status_details_callback = StatusDetailsCallback(
            training=self,
            status="TRAINING",
            training_round_step_id=training_round_step_id,
        )

        if base_model_repo_id == "spicecloud/spice-cnn-base":
            training_args.remove_unused_columns = False

        trainer = Trainer(
            model=model,
            args=training_args,
            train_dataset=tokenized_dataset,
            tokenizer=tokenizer,
            compute_metrics=compute_metrics,
            callbacks=[status_details_callback],
            data_collator=DefaultDataCollator(),
        )

        self._update_training_round_step(
            training_round_step_id=training_round_step_id, status="TRAINING"
        )

        trainer.train()

        trainer.save_model(output_dir=str(model_cache_for_training_round))

        self._update_training_round_step(
            training_round_step_id=training_round_step_id, status="TRAINING_COMPLETE"
        )

        # clear the cache
        train_dataset.cleanup_cache_files()

    def test_model(self):
        config = read_config_file(filepath=SPICE_TRAINING_FILEPATH)
        training_round_step_id = config["id"]
        if not self.check_training_round_step_exists(training_round_step_id):
            return None

        training_round_id = config["trainingRound"]["id"]
        base_model_repo_id = config["trainingRound"]["trainingJob"]["baseModel"][
            "repoId"
        ]
        # create the folder for the new training round
        # where the step model will be saved
        training_round_directory = f"{training_round_id}/steps/{training_round_step_id}"
        model_cache_for_training_round = SPICE_MODEL_CACHE_FILEPATH.joinpath(
            training_round_directory
        )

        # get dataset
        print("Loading test dataset...")
        self._update_training_round_step(
            training_round_step_id=training_round_step_id,
            status="DOWNLOADING_TESTING_DATASET",
        )
        test_dataset = self._load_dataset(config, "test")

        # tokenize dataset
        tokenizer = None
        if base_model_repo_id == "spicecloud/spice-cnn-base":
            tokenizer, tokenized_dataset = self._process_dataset(
                test_dataset, config, "test"
            )
        elif base_model_repo_id == "bert-base-uncased":
            tokenizer, tokenized_dataset = self._tokenize_dataset(
                test_dataset, config, "test"
            )
        else:
            error_message = (
                f"test_model has unknown base model: {base_model_repo_id} type"
            )
            LOGGER.error(error_message)
            raise ValueError(error_message)

        # load your model with the number of expected labels:
        model = self._load_model(config, "test")

        eval_args = self._get_training_arguments(
            model_cache_for_training_round, config, "test"
        )

        metric = evaluate.load("accuracy")

        def compute_metrics(eval_pred):
            logits, labels = eval_pred
            predictions = np.argmax(logits, axis=-1)
            return metric.compute(predictions=predictions, references=labels)

        status_details_callback = StatusDetailsCallback(
            training=self,
            training_round_step_id=training_round_step_id,
            status="TESTING",
        )

        if base_model_repo_id == "spicecloud/spice-cnn-base":
            eval_args.remove_unused_columns = False

        trainer = Trainer(
            model=model,
            args=eval_args,
            eval_dataset=tokenized_dataset,
            tokenizer=tokenizer,
            compute_metrics=compute_metrics,
            callbacks=[status_details_callback],
            data_collator=DefaultDataCollator(),
        )

        self._update_training_round_step(
            training_round_step_id=training_round_step_id,
            status="TESTING",
        )

        metrics = trainer.evaluate()

        trainer.log_metrics("eval", metrics)
        trainer.save_metrics("eval", metrics, combined=False)

        self._update_training_round_step(
            training_round_step_id=training_round_step_id,
            status="TESTING_COMPLETE",
            metrics=metrics,
        )

        # clear the cache
        test_dataset.cleanup_cache_files()

    def _download_verify_model_file(self, url, destination_url, is_json=False):
        response = requests.get(url)
        response.raise_for_status()

        # Implement checksum if file is dated
        if is_json:
            with open(destination_url, "w") as file:  # download json file
                json_string = response.content.decode("utf-8")
                json_data = json.loads(json_string)
                # TEMPORARY FORMATTING FIX THIS SHOULD BE DONE IN UPLOADER
                json_formatted = json.dumps(json_data, indent=2) + "\n"
                file.write(json_formatted)
        else:
            with open(destination_url, "wb") as file:  # download binary file
                file.write(response.content)

    def check_training_round_exists(self, training_round_id) -> bool:
        try:
            result = self._get_training_round(training_round_id=training_round_id)
            if (
                result["trainingRound"]["status"]
                not in self.spice.worker.ACTIVE_STATUSES
            ):
                LOGGER.info(
                    f"Training round {training_round_id} is not active. Deleting {SPICE_ROUND_VERIFICATION_FILEPATH} config."  # noqa
                )
                SPICE_ROUND_VERIFICATION_FILEPATH.unlink(missing_ok=True)
                return False
        except TransportQueryError as exception:
            if exception.errors:
                for error in exception.errors:
                    if (
                        error.get("message", "")
                        == "TrainingRound matching query does not exist."
                    ):
                        LOGGER.error(
                            f""" [*] Training Round ID: {training_round_id} not found. Deleting {SPICE_ROUND_VERIFICATION_FILEPATH} config."""  # noqa
                        )
                        SPICE_ROUND_VERIFICATION_FILEPATH.unlink(missing_ok=True)
                        return False
                raise exception
            else:
                raise exception
        return True

    def check_training_round_step_exists(self, training_round_step_id) -> bool:
        try:
            result = self._get_training_round_step(
                training_round_step_id=training_round_step_id
            )
            if (
                result["trainingRoundStep"]["status"]
                not in self.spice.worker.ACTIVE_STATUSES
            ):
                LOGGER.info(
                    f"Training round step {training_round_step_id} is not active. Deleting {SPICE_TRAINING_FILEPATH} config."  # noqa
                )
                SPICE_TRAINING_FILEPATH.unlink(missing_ok=True)
                return False
        except TransportQueryError as exception:
            if exception.errors:
                for error in exception.errors:
                    if (
                        error.get("message", "")
                        == "TrainingRoundStep matching query does not exist."
                    ):
                        LOGGER.error(
                            f""" [*] Training Round Step ID: {training_round_step_id} not found. Deleting {SPICE_TRAINING_FILEPATH}."""  # noqa
                        )
                        SPICE_TRAINING_FILEPATH.unlink(missing_ok=True)
                        return False
                raise exception
            else:
                raise exception
        return True

    def verify_model(self):
        config = read_config_file(filepath=SPICE_ROUND_VERIFICATION_FILEPATH)
        training_round_id = config["id"]

        if not self.check_training_round_exists(training_round_id=training_round_id):
            return None

        training_round_number = config["roundNumber"]
        training_job_id = config["trainingJob"]["id"]
        base_model_repo_id = config["trainingJob"]["baseModelRepoId"]

        # create the folder for the verification round
        # where the step model will be saved
        verification_round_directory = f"{training_round_id}/verification"
        verification_cache_for_training_round = SPICE_MODEL_CACHE_FILEPATH.joinpath(
            verification_round_directory
        )

        # get the presigned urls for a round's model + configs
        result = self._get_agent_round_presigned_urls(
            training_round_id=training_round_id,
        )

        agent_round_presigned_round_model_url = result["getAgentRoundPresignedUrls"][
            "roundModel"
        ]
        agent_presigned_config_url = result["getAgentRoundPresignedUrls"]["config"]

        destination_round_file_url = SPICE_MODEL_CACHE_FILEPATH.joinpath(
            f"{training_round_id}/pytorch_model.bin"
        )

        destination_config_url = SPICE_MODEL_CACHE_FILEPATH.joinpath(
            f"{training_round_id}/config.json"
        )

        print("Downloading round model...")
        self._update_training_round(
            training_round_id=training_round_id, status="DOWNLOADING_ROUND_MODEL"
        )
        self._download_verify_model_file(
            agent_round_presigned_round_model_url, destination_round_file_url
        )

        self._download_verify_model_file(
            agent_presigned_config_url, destination_config_url, is_json=True
        )

        # download validation dataset
        print("Loading verification dataset...")
        self._update_training_round(
            training_round_id=training_round_id,
            status="DOWNLOADING_VERIFICATION_DATASET",
        )
        test_dataset = self._load_dataset(config, "verify")

        # tokenize dataset
        tokenizer = None
        if base_model_repo_id == "spicecloud/spice-cnn-base":
            tokenizer, tokenized_dataset = self._process_dataset(
                test_dataset, config, "verify"
            )
        elif base_model_repo_id == "bert-base-uncased":
            tokenizer, tokenized_dataset = self._tokenize_dataset(
                test_dataset, config, "verify"
            )
        else:
            error_message = (
                f"test_model has unknown base model: {base_model_repo_id} type"
            )
            LOGGER.error(error_message)
            raise ValueError(error_message)

        # load your model with the number of expected labels:
        model = self._load_model(config, "verify")

        eval_args = self._get_training_arguments(
            verification_cache_for_training_round, config, "verify"
        )

        metric = evaluate.load("accuracy")

        def compute_metrics(eval_pred):
            logits, labels = eval_pred
            predictions = np.argmax(logits, axis=-1)
            return metric.compute(predictions=predictions, references=labels)

        status_details_callback = StatusDetailsCallback(
            training=self,
            training_round_id=training_round_id,
            status="VERIFYING",
        )

        if base_model_repo_id == "spicecloud/spice-cnn-base":
            eval_args.remove_unused_columns = False

        trainer = Trainer(
            model=model,
            args=eval_args,
            eval_dataset=tokenized_dataset,
            tokenizer=tokenizer,
            compute_metrics=compute_metrics,
            callbacks=[status_details_callback],
            data_collator=DefaultDataCollator(),
        )

        self._update_training_round(
            training_round_id=training_round_id,
            status="VERIFYING",
        )
        metrics = trainer.evaluate()

        trainer.log_metrics("eval", metrics)
        trainer.save_metrics("eval", metrics, combined=False)

        # Upload evaluation results
        bucket_dir = f"{training_job_id}/rounds/{training_round_number}/merged/"
        for file in verification_cache_for_training_round.iterdir():
            if file.is_file():
                bucket_key = bucket_dir + file.name

                # create the "file" and attach it to the training round step
                file_checksum = hashlib.md5(file.read_bytes()).hexdigest()
                file_id = self.spice.uploader._create_file(
                    file_name=file.name,
                    file_size=file.stat().st_size,
                    file_checksum=file_checksum,
                    location=f"s3://{bucket_key}",
                )

                self.spice.uploader.upload_file_direct(
                    bucket_name=MODEL_BUCKET_NAME,
                    key=bucket_key,
                    filepath=file,
                    file_id=file_id,
                )

        print("Complete!")

        self._update_training_round(
            training_round_id=training_round_id,
            status="VERIFYING_COMPLETE",
            metrics=metrics,
        )

        # clear the cache
        test_dataset.cleanup_cache_files()



from typing import Dict

from gql import Client
from gql.transport.aiohttp import AIOHTTPTransport
import requests

from spice_agent.__version__ import __version__


def create_session(
    host_config: Dict,
    host: str = "api.spice.cloud",
    fetch_schema_from_transport: bool = False,
):
    token = host_config.get("token")
    transport = host_config.get("transport")
    url = f"{transport}://{host}/"
    headers = {
        "Content-Type": "application/json",
        "Accept": "application/json",
        "x-spice-agent": f"spice@{__version__}",
    }
    if token:
        headers["Authorization"] = f"Token {token}"

    if fingerprint := host_config.get("fingerprint", None):
        headers["x-spice-fingerprint"] = fingerprint
    transport = AIOHTTPTransport(url=url, headers=headers)
    session = Client(
        transport=transport,
        fetch_schema_from_transport=fetch_schema_from_transport,
    )
    return session


def create_requests_session(
    host_config: Dict,
):
    token = host_config.get("token")

    headers = {
        # "Content-Type": "multipart/form-data", # DO NOT ADD THIS MIME TYPE IT BREAKS
        "Accept": "application/json",
        "x-spice-agent": f"spice@{__version__}",
    }
    if token:
        headers["Authorization"] = f"Token {token}"

    if fingerprint := host_config.get("fingerprint", None):
        headers["x-spice-fingerprint"] = fingerprint

    session = requests.Session()
    session.headers.update(headers)
    return session



import json

import click


def print_result(message: str, context: click.Context, fg: str = None):
    """Print message to stdout or stderr"""
    if context.obj["JSON"]:
        message = json.dumps(message, indent=2)
    click.secho(message, fg=fg)



import os
import configparser
import subprocess
from pathlib import Path

HOME_DIRECTORY = Path.home()

SPICE_BINARY_PATH = Path(HOME_DIRECTORY / ".pyenv" / "shims" / "spice")

SPICE_AGENT_SERVICE = "cloud.spice.agent.service"
SPICE_AGENT_LOGS = "cloud.spice.agent.log"
SPICE_AGENT_SERVICE_FILEPATH = Path(
    HOME_DIRECTORY / ".config" / "systemd" / "user" / SPICE_AGENT_SERVICE
)
SPICE_AGENT_LOGS_FILEPATH = Path(HOME_DIRECTORY / ".cache" / "spice" / SPICE_AGENT_LOGS)


def stop_service():
    try:
        if is_service_active():
            stop_existing_service = f"systemctl --user stop {SPICE_AGENT_SERVICE}"
            subprocess.check_output(stop_existing_service.split(" "))
        return True
    except subprocess.CalledProcessError as exception:
        print("stop_service: ", exception)
        return False


def is_service_active():
    try:
        is_service_active = (
            f"systemctl --user show {SPICE_AGENT_SERVICE} -p ActiveState --value"
        )
        output = subprocess.check_output(is_service_active.split(" ")).decode().strip()
        return output == "active"
    except subprocess.CalledProcessError as exception:
        print("is_service_active: ", exception)
        return False


def disable_service():
    try:
        if is_service_enabled():
            disable_existing_service = f"systemctl --user disable {SPICE_AGENT_SERVICE}"
            subprocess.check_output(disable_existing_service.split(" "))
        return True
    except subprocess.CalledProcessError as exception:
        print("disable_service: ", exception)
        return False


def is_service_enabled():
    try:
        is_service_active = f"systemctl --user show {SPICE_AGENT_SERVICE} -p UnitFileState --value"  # noqa
        output = subprocess.check_output(is_service_active.split(" ")).decode().strip()
        return output == "enabled"
    except subprocess.CalledProcessError as exception:
        print("is_service_enabled: ", exception)
        return False


def populate_service_file():
    SPICE_AGENT_LOGS_FILEPATH.parent.mkdir(parents=True, exist_ok=True)
    SPICE_AGENT_LOGS_FILEPATH.touch()

    if SPICE_AGENT_SERVICE_FILEPATH.exists():
        SPICE_AGENT_SERVICE_FILEPATH.unlink()
    SPICE_AGENT_SERVICE_FILEPATH.parent.mkdir(parents=True, exist_ok=True)
    SPICE_AGENT_SERVICE_FILEPATH.touch()

    config = configparser.ConfigParser()
    config.optionxform = str  # type: ignore

    config["Unit"] = {
        "Description": "Spice Agent",
        "After": "network.target",
    }

    config["Service"] = {
        "ExecStart": f"{SPICE_BINARY_PATH} worker start",
        "Restart": "always",
        "StandardError": f"append:{SPICE_AGENT_LOGS_FILEPATH}",
        "StandardOutput": f"append:{SPICE_AGENT_LOGS_FILEPATH}",
    }

    config["Install"] = {
        "WantedBy": "default.target",
    }

    try:
        with open(SPICE_AGENT_SERVICE_FILEPATH, "w") as service_file:
            config.write(service_file)
    except IOError as exception:
        print(f"populate_service_file: {exception}")

    SPICE_AGENT_SERVICE_FILEPATH.chmod(0o644)
    verify_service_file()


def verify_service_file():
    systemctl_check = f"systemctl --user show {SPICE_AGENT_SERVICE} -p CanStart --value"
    try:
        output = subprocess.check_output(systemctl_check.split(" ")).decode().strip()
        if output == "no":
            raise Exception(f"{systemctl_check} yielded {output}")
        return True
    except subprocess.CalledProcessError as exception:
        print("verify_service_file: ", exception)
        return False


def create_stdout_file():
    if not SPICE_AGENT_LOGS_FILEPATH.exists():
        SPICE_AGENT_LOGS_FILEPATH.parent.mkdir(parents=True, exist_ok=True)
        SPICE_AGENT_LOGS_FILEPATH.touch()


def delete_stdout_file():
    if SPICE_AGENT_LOGS_FILEPATH.exists():
        SPICE_AGENT_LOGS_FILEPATH.unlink()


def enable_service():
    try:
        enable_service = f"systemctl --user enable {SPICE_AGENT_SERVICE}"
        subprocess.check_output(enable_service.split(" "))
        return True
    except subprocess.CalledProcessError as exception:
        print("enable_service: ", exception)
        return False


def start_service():
    try:
        start_new_service = f"systemctl --user start {SPICE_AGENT_SERVICE}"
        subprocess.check_output(start_new_service.split(" "))
        return True
    except subprocess.CalledProcessError as exception:
        print("start_service: ", exception)
        return False


def view_service_logs():
    follow_logs = f"tail -f -n +1 {SPICE_AGENT_LOGS_FILEPATH}"
    os.system(follow_logs)


def full_service_install():
    stop_service()
    disable_service()
    populate_service_file()
    create_stdout_file()
    enable_service()
    start_service()


def full_service_uninstall():
    stop_service()
    disable_service()
    if SPICE_AGENT_SERVICE_FILEPATH.exists():
        SPICE_AGENT_SERVICE_FILEPATH.unlink()
    delete_stdout_file()



import os
from pathlib import Path
import subprocess

HOME_DIRECTORY = Path.home()
CURRENT_USER = str(HOME_DIRECTORY.expanduser()).lstrip("/Users/")

SPICE_BINARY_PATH = Path(HOME_DIRECTORY / ".pyenv" / "shims" / "spice")
SPICE_LAUNCH_AGENT_FILEPATH = Path(
    HOME_DIRECTORY / "Library" / "LaunchAgents" / "cloud.spice.agent.plist"
)
SPICE_LAUNCH_AGENT_LOGS = Path(
    HOME_DIRECTORY / "Library" / "Logs" / "cloud.spice.agent.log"
)
SPICE_LAUNCH_AGENT_LABEL = "cloud.spice.agent"
SPICE_LAUNCH_AGENT_WORKING_DIR = Path(HOME_DIRECTORY / "Logs" / "cloud.spice.agent.log")


def stop_launch_agent():
    try:
        stop_existing_process = f"launchctl stop {SPICE_LAUNCH_AGENT_LABEL}"
        subprocess.check_output(stop_existing_process.split(" "))
        return True
    except subprocess.CalledProcessError as exception:
        print("stop_launch_agent: ", exception)
        return False


def start_launch_agent():
    try:
        start_new_agent = f"launchctl start {SPICE_LAUNCH_AGENT_LABEL}"
        subprocess.check_output(start_new_agent.split(" "))
        return True
    except subprocess.CalledProcessError as exception:
        print("start_launch_agent: ", exception)
        return False


def unload_launch_agent():
    try:
        remove_existing_agent = f"launchctl unload -w {SPICE_LAUNCH_AGENT_FILEPATH}"
        subprocess.check_output(remove_existing_agent.split(" "))
        return True
    except subprocess.CalledProcessError as exception:
        print("remove_launch_agent: ", exception)
        return False


def load_launch_agent():
    try:
        load_new_plist = f"launchctl load -w {SPICE_LAUNCH_AGENT_FILEPATH}"
        subprocess.check_output(load_new_plist.split(" "))
        return True
    except subprocess.CalledProcessError as exception:
        print("load_launch_agent: ", exception)
        return False


def create_stdout_file():
    if not SPICE_LAUNCH_AGENT_LOGS.exists():
        SPICE_LAUNCH_AGENT_LOGS.parent.mkdir(parents=True, exist_ok=True)
        SPICE_LAUNCH_AGENT_LOGS.touch()


def delete_stdout_file():
    if SPICE_LAUNCH_AGENT_LOGS.exists():
        SPICE_LAUNCH_AGENT_LOGS.unlink()


def populate_fresh_launch_agent():
    SPICE_LAUNCH_AGENT_LOGS.parent.mkdir(parents=True, exist_ok=True)
    SPICE_LAUNCH_AGENT_LOGS.touch()

    if SPICE_LAUNCH_AGENT_FILEPATH.exists():
        SPICE_LAUNCH_AGENT_FILEPATH.unlink()
    SPICE_LAUNCH_AGENT_FILEPATH.parent.mkdir(parents=True, exist_ok=True)
    SPICE_LAUNCH_AGENT_FILEPATH.touch()

    SPICE_LAUNCH_AGENT_FILEPATH.write_text(
        f"""<?xml version="1.0" encoding="UTF-8"?>
<!DOCTYPE plist PUBLIC "-//Apple//DTD PLIST 1.0//EN" "http://www.apple.com/DTDs/PropertyList-1.0.dtd">
<plist version="1.0">
<dict>
    <key>KeepAlive</key>
    <true/>
    <key>Label</key>
    <string>{SPICE_LAUNCH_AGENT_LABEL}</string>
    <key>LimitLoadToSessionType</key>
	<array>
		<string>Aqua</string>
		<string>Background</string>
		<string>LoginWindow</string>
		<string>StandardIO</string>
	</array>
    <key>ProgramArguments</key>
    <array>
        <string>{SPICE_BINARY_PATH}</string>
        <string>worker</string>
        <string>start</string>
    </array>
    <key>RunAtLoad</key>
    <true/>
    <key>StandardErrorPath</key>
    <string>{SPICE_LAUNCH_AGENT_LOGS}</string>
    <key>StandardOutPath</key>
    <string>{SPICE_LAUNCH_AGENT_LOGS}</string>
</dict>
</plist>"""  # noqa: E501
    )
    SPICE_LAUNCH_AGENT_FILEPATH.chmod(0o644)
    verify_launch_agent()


def verify_launch_agent():
    plutil_check = f"plutil -lint {SPICE_LAUNCH_AGENT_FILEPATH}"
    try:
        subprocess.check_output(plutil_check.split(" "))
        return True
    except subprocess.CalledProcessError as exception:
        print("verify_launch_agent: ", exception)
        return False


def view_launch_agent_logs():
    follow_logs = f"tail -f -n +1 {SPICE_LAUNCH_AGENT_LOGS}"
    os.system(follow_logs)


def full_launch_agent_install():
    stop_launch_agent()
    unload_launch_agent()
    populate_fresh_launch_agent()
    create_stdout_file()
    load_launch_agent()
    start_launch_agent()


def full_launch_agent_uninstall():
    stop_launch_agent()
    unload_launch_agent()
    if SPICE_LAUNCH_AGENT_FILEPATH.exists():
        SPICE_LAUNCH_AGENT_FILEPATH.unlink()
    delete_stdout_file()



import os
import sys

import click

from spice_agent.__version__ import __version__
from spice_agent.client import Spice

from .auth.commands import config_command, whoami_command
from .daemons.commands import cli as daemons_cli
from .hardware.commands import cli as hardware_cli
from .inference.commands import cli as inference_cli
from .training.commands import cli as training_cli
from .worker.commands import cli as worker_cli


@click.group()
@click.option(
    "--host",
    required=False,
    default=lambda: os.environ.get("SPICE_HOST", "api.spice.cloud"),
    show_default="api.spice.cloud",
    help="Environment of Spice API",
)
@click.option(
    "--yes", is_flag=True, show_default=True, default=False, help="Skip interactions."
)
@click.option(
    "--debug", is_flag=True, show_default=True, default=False, help="Enable debug mode."
)
@click.option(
    "--json",
    is_flag=True,
    show_default=True,
    default=False,
    help="Turn all outputs into JSON.",
)
@click.option(
    "-v",
    "--verbose",
    count=True,
    help="Verbosity of output levels.",
)
@click.version_option(__version__)
@click.pass_context
def cli(context, host, yes, debug, json, verbose):
    """spice agent - a CLI tool for https://spice.cloud"""
    context.ensure_object(dict)
    context.obj["YES"] = yes
    context.obj["DEBUG"] = debug
    context.obj["JSON"] = json
    context.obj["VERBOSE"] = verbose
    context.obj["HOST"] = host
    if "config" not in sys.argv:
        context.obj["SPICE"] = Spice(host=host, DEBUG=debug)


cli.add_command(config_command, "config")
cli.add_command(whoami_command, "whoami")
cli.add_command(hardware_cli, "hardware")
cli.add_command(inference_cli, "inference")
cli.add_command(training_cli, "training")
cli.add_command(daemons_cli, "daemon")
cli.add_command(worker_cli, "worker")

if __name__ == "__main__":
    cli(obj={})



import torch
from typing import Optional, Dict, Union, List, Any, Tuple
from dataclasses import dataclass
from PIL import Image as PILImage
import numpy as np


# Data Classes for Stable Diffusion Pipeline -------------------------------------------
@dataclass
class InputForStableDiffusionPipeline:
    """
    Inputs for Stable Diffusion Pipeline

    Args:
        prompt (`str` or `List[str]`, *optional*):
            The prompt or prompts to guide the image generation. If not defined, one has to pass `prompt_embeds`.
            instead.
        negative_prompt (`str` or `List[str]`, *optional*):
            The prompt or prompts not to guide the image generation. If not defined, one has to pass
            `negative_prompt_embeds` instead. Ignored when not using guidance (i.e., ignored if `guidance_scale` is
            less than `1`).
        prompt_embeds (`torch.FloatTensor`, *optional*):
            Pre-generated text embeddings. Can be used to easily tweak text inputs, *e.g.* prompt weighting. If not
            provided, text embeddings will be generated from `prompt` input argument.
        negative_prompt_embeds (`torch.FloatTensor`, *optional*):
            Pre-generated negative text embeddings. Can be used to easily tweak text inputs, *e.g.* prompt
            weighting. If not provided, negative_prompt_embeds will be generated from `negative_prompt` input
            argument.
    """  # noqa

    prompt: Optional[Union[str, List[str]]] = None
    negative_prompt: Optional[Union[str, List[str]]] = None
    prompt_embeds: Optional[torch.FloatTensor] = None
    negative_prompt_embeds: Optional[torch.FloatTensor] = None


@dataclass
class InferenceOptionsForStableDiffusionPipeline:
    """
    Inference options for Stable Diffusion Pipeline

    Args:
        height (`int`, *optional*, defaults to self.unet.config.sample_size * self.vae_scale_factor):
            The height in pixels of the generated image.
        width (`int`, *optional*, defaults to self.unet.config.sample_size * self.vae_scale_factor):
            The width in pixels of the generated image.
        num_inference_steps (`int`, *optional*, defaults to 50):
            The number of denoising steps. More denoising steps usually lead to a higher quality image at the
            expense of slower inference.
        guidance_scale (`float`, *optional*, defaults to 7.5):
            Guidance scale as defined in [Classifier-Free Diffusion Guidance](https://arxiv.org/abs/2207.12598).
            `guidance_scale` is defined as `w` of equation 2. of [Imagen
            Paper](https://arxiv.org/pdf/2205.11487.pdf). Guidance scale is enabled by setting `guidance_scale >
            1`. Higher guidance scale encourages to generate images that are closely linked to the text `prompt`,
            usually at the expense of lower image quality.
        num_images_per_prompt (`int`, *optional*, defaults to 1):
            The number of images to generate per prompt.
        eta (`float`, *optional*, defaults to 0.0):
            Corresponds to parameter eta (η) in the DDIM paper: https://arxiv.org/abs/2010.02502. Only applies to
            [`schedulers.DDIMScheduler`], will be ignored for others.
        callback_steps (`int`, *optional*, defaults to 1):
            The frequency at which the `callback` function will be called. If not specified, the callback will be
            called at every step.
        cross_attention_kwargs (`dict`, *optional*):
            A kwargs dictionary that if specified is passed along to the `AttentionProcessor` as defined under
            `self.processor` in
            [diffusers.cross_attention](https://github.com/huggingface/diffusers/blob/main/src/diffusers/models/cross_attention.py).
        guidance_rescale (`float`, *optional*, defaults to 0.7):
            Guidance rescale factor proposed by [Common Diffusion Noise Schedules and Sample Steps are
            Flawed](https://arxiv.org/pdf/2305.08891.pdf) `guidance_scale` is defined as `φ` in equation 16. of
            [Common Diffusion Noise Schedules and Sample Steps are Flawed](https://arxiv.org/pdf/2305.08891.pdf).
            Guidance rescale factor should fix overexposure when using zero terminal SNR.
    """  # noqa

    height: Optional[int] = None
    width: Optional[int] = None
    num_inference_steps: int = 50
    guidance_scale: float = 7.5
    num_images_per_prompt: Optional[int] = 1
    eta: float = 0.0
    callback_steps: int = 1
    cross_attention_kwargs: Optional[Dict[str, Any]] = None
    guidance_rescale: float = 0.7


@dataclass
class OutputForStableDiffusionPipeline:
    """
    Output related arguments for Stable Diffusion Pipeline

    Args:
        generator (`torch.Generator` or `List[torch.Generator]`, *optional*):
            One or a list of [torch generator(s)](https://pytorch.org/docs/stable/generated/torch.Generator.html)
            to make generation deterministic.
        latents (`torch.FloatTensor`, *optional*):
            Pre-generated noisy latents, sampled from a Gaussian distribution, to be used as inputs for image
            generation. Can be used to tweak the same generation with different prompts. If not provided, a latents
            tensor will ge generated by sampling using the supplied random `generator`.
        output_type (`str`, *optional*, defaults to `"pil"`):
            The output format of the generate image. Choose between
            [PIL](https://pillow.readthedocs.io/en/stable/): `PIL.Image.Image` or `np.array`.
        return_dict (`bool`, *optional*, defaults to `True`):
            Whether or not to return a [`~pipelines.stable_diffusion.StableDiffusionPipelineOutput`] instead of a
            plain tuple.
    """  # noqa

    # generator: Optional[Union[torch.Generator, List[torch.Generator]]] = None
    latents: Optional[torch.FloatTensor] = None
    output_type: Optional[str] = "pil"
    return_dict: bool = True


# Data Classes for Stable Diffusion XL Pipeline ----------------------------------------
@dataclass
class InputForStableDiffusionXLPipeline:
    """
    Inputs for Stable Diffusion XL Pipeline

    Args:
        prompt (`str` or `List[str]`, *optional*):
            The prompt or prompts to guide the image generation. If not defined, one has to pass `prompt_embeds`.
            instead.
        negative_prompt (`str` or `List[str]`, *optional*):
            The prompt or prompts not to guide the image generation. If not defined, one has to pass
            `negative_prompt_embeds` instead. Ignored when not using guidance (i.e., ignored if `guidance_scale` is
            less than `1`).
        prompt_embeds (`torch.FloatTensor`, *optional*):
            Pre-generated text embeddings. Can be used to easily tweak text inputs, *e.g.* prompt weighting. If not
            provided, text embeddings will be generated from `prompt` input argument.
        negative_prompt_embeds (`torch.FloatTensor`, *optional*):
            Pre-generated negative text embeddings. Can be used to easily tweak text inputs, *e.g.* prompt
            weighting. If not provided, negative_prompt_embeds will be generated from `negative_prompt` input
            argument.
        pooled_prompt_embeds (`torch.FloatTensor`, *optional*):
                Pre-generated pooled text embeddings. Can be used to easily tweak text inputs, *e.g.* prompt weighting.
                If not provided, pooled text embeddings will be generated from `prompt` input argument.
        negative_pooled_prompt_embeds (`torch.FloatTensor`, *optional*):
            Pre-generated negative pooled text embeddings. Can be used to easily tweak text inputs, *e.g.* prompt
            weighting. If not provided, pooled negative_prompt_embeds will be generated from `negative_prompt`
            input argument.
    """  # noqa

    prompt: Optional[Union[str, List[str]]] = None
    negative_prompt: Optional[Union[str, List[str]]] = None
    prompt_embeds: Optional[torch.FloatTensor] = None
    negative_prompt_embeds: Optional[torch.FloatTensor] = None
    pooled_prompt_embeds: Optional[torch.FloatTensor] = None
    negative_pooled_prompt_embeds: Optional[torch.FloatTensor] = None


@dataclass
class InferenceOptionsForStableDiffusionXLPipeline:
    """
    Inference Options for Stable Diffusion XL Pipeline

    Args:
        height (`int`, *optional*, defaults to self.unet.config.sample_size * self.vae_scale_factor):
            The height in pixels of the generated image.
        width (`int`, *optional*, defaults to self.unet.config.sample_size * self.vae_scale_factor):
            The width in pixels of the generated image.
        num_inference_steps (`int`, *optional*, defaults to 50):
            The number of denoising steps. More denoising steps usually lead to a higher quality image at the
            expense of slower inference.
        guidance_scale (`float`, *optional*, defaults to 7.5):
            Guidance scale as defined in [Classifier-Free Diffusion Guidance](https://arxiv.org/abs/2207.12598).
            `guidance_scale` is defined as `w` of equation 2. of [Imagen
            Paper](https://arxiv.org/pdf/2205.11487.pdf). Guidance scale is enabled by setting `guidance_scale >
            1`. Higher guidance scale encourages to generate images that are closely linked to the text `prompt`,
            usually at the expense of lower image quality.
        num_images_per_prompt (`int`, *optional*, defaults to 1):
            The number of images to generate per prompt.
        eta (`float`, *optional*, defaults to 0.0):
            Corresponds to parameter eta (η) in the DDIM paper: https://arxiv.org/abs/2010.02502. Only applies to
            [`schedulers.DDIMScheduler`], will be ignored for others.
        callback_steps (`int`, *optional*, defaults to 1):
            The frequency at which the `callback` function will be called. If not specified, the callback will be
            called at every step.
        cross_attention_kwargs (`dict`, *optional*):
            A kwargs dictionary that if specified is passed along to the `AttentionProcessor` as defined under
            `self.processor` in
            [diffusers.cross_attention](https://github.com/huggingface/diffusers/blob/main/src/diffusers/models/cross_attention.py).
        guidance_rescale (`float`, *optional*, defaults to 0.7):
            Guidance rescale factor proposed by [Common Diffusion Noise Schedules and Sample Steps are
            Flawed](https://arxiv.org/pdf/2305.08891.pdf) `guidance_scale` is defined as `φ` in equation 16. of
            [Common Diffusion Noise Schedules and Sample Steps are Flawed](https://arxiv.org/pdf/2305.08891.pdf).
            Guidance rescale factor should fix overexposure when using zero terminal SNR.
        original_size (`Tuple[int]`, *optional*, defaults to (1024, 1024)):
            TODO
        crops_coords_top_left (`Tuple[int]`, *optional*, defaults to (0, 0)):
            TODO
        target_size (`Tuple[int]`, *optional*, defaults to (1024, 1024)):
            TODO
    """  # noqa

    height: Optional[int] = None
    width: Optional[int] = None
    num_inference_steps: int = 50
    guidance_scale: float = 7.5
    num_images_per_prompt: Optional[int] = 1
    eta: float = 0.0
    callback_steps: int = 1
    cross_attention_kwargs: Optional[Dict[str, Any]] = None
    guidance_rescale: float = 0.7
    original_size: Optional[Tuple[int, int]] = None
    crops_coords_top_left: Tuple[int, int] = (0, 0)
    target_size: Optional[Tuple[int, int]] = None


# Data Classes for Stable Diffusion XL IMG2IMG Pipeline --------------------------------


@dataclass
class InputForStableDiffusionXLImg2ImgPipeline:
    """
    Input for Stable Diffusion XL Img2Img Pipeline

    Args:
        image (`torch.FloatTensor` or `PIL.Image.Image` or `np.ndarray` or `List[torch.FloatTensor]` or `List[PIL.Image.Image]` or `List[np.ndarray]`):
            The image(s) to modify with the pipeline.
        prompt (`str` or `List[str]`, *optional*):
            The prompt or prompts to guide the image generation. If not defined, one has to pass `prompt_embeds`.
            instead.
        negative_prompt (`str` or `List[str]`, *optional*):
            The prompt or prompts not to guide the image generation. If not defined, one has to pass
            `negative_prompt_embeds` instead. Ignored when not using guidance (i.e., ignored if `guidance_scale` is
            less than `1`).
        prompt_embeds (`torch.FloatTensor`, *optional*):
            Pre-generated text embeddings. Can be used to easily tweak text inputs, *e.g.* prompt weighting. If not
            provided, text embeddings will be generated from `prompt` input argument.
        negative_prompt_embeds (`torch.FloatTensor`, *optional*):
            Pre-generated negative text embeddings. Can be used to easily tweak text inputs, *e.g.* prompt
            weighting. If not provided, negative_prompt_embeds will be generated from `negative_prompt` input
            argument.
        pooled_prompt_embeds (`torch.FloatTensor`, *optional*):
            Pre-generated pooled text embeddings. Can be used to easily tweak text inputs, *e.g.* prompt weighting.
            If not provided, pooled text embeddings will be generated from `prompt` input argument.
        negative_pooled_prompt_embeds (`torch.FloatTensor`, *optional*):
            Pre-generated negative pooled text embeddings. Can be used to easily tweak text inputs, *e.g.* prompt
            weighting. If not provided, pooled negative_prompt_embeds will be generated from `negative_prompt`
            input argument.
    """  # noqa

    image: Optional[
        Union[
            torch.FloatTensor,
            PILImage.Image,
            np.ndarray,
            List[torch.FloatTensor],
            List[PILImage.Image],
            List[np.ndarray],
        ]
    ] = None

    prompt: Optional[Union[str, List[str]]] = None
    negative_prompt: Optional[Union[str, List[str]]] = None
    prompt_embeds: Optional[torch.FloatTensor] = None
    negative_prompt_embeds: Optional[torch.FloatTensor] = None
    pooled_prompt_embeds: Optional[torch.FloatTensor] = None
    negative_pooled_prompt_embeds: Optional[torch.FloatTensor] = None


@dataclass
class InferenceOptionsForStableDiffusionXLImg2ImgPipeline:
    """
    Args:
        strength (`float`, *optional*, defaults to 0.8):
            Conceptually, indicates how much to transform the reference `image`. Must be between 0 and 1. `image`
            will be used as a starting point, adding more noise to it the larger the `strength`. The number of
            denoising steps depends on the amount of noise initially added. When `strength` is 1, added noise will
            be maximum and the denoising process will run for the full number of iterations specified in
            `num_inference_steps`. A value of 1, therefore, essentially ignores `image`.
        num_inference_steps (`int`, *optional*, defaults to 50):
            The number of denoising steps. More denoising steps usually lead to a higher quality image at the
            expense of slower inference.
        guidance_scale (`float`, *optional*, defaults to 7.5):
            Guidance scale as defined in [Classifier-Free Diffusion Guidance](https://arxiv.org/abs/2207.12598).
            `guidance_scale` is defined as `w` of equation 2. of [Imagen
            Paper](https://arxiv.org/pdf/2205.11487.pdf). Guidance scale is enabled by setting `guidance_scale >
            1`. Higher guidance scale encourages to generate images that are closely linked to the text `prompt`,
            usually at the expense of lower image quality.
        num_images_per_prompt (`int`, *optional*, defaults to 1):
            The number of images to generate per prompt.
        eta (`float`, *optional*, defaults to 0.0):
            Corresponds to parameter eta (η) in the DDIM paper: https://arxiv.org/abs/2010.02502. Only applies to
            [`schedulers.DDIMScheduler`], will be ignored for others.
        callback_steps (`int`, *optional*, defaults to 1):
            The frequency at which the `callback` function will be called. If not specified, the callback will be
            called at every step.
        cross_attention_kwargs (`dict`, *optional*):
            A kwargs dictionary that if specified is passed along to the `AttentionProcessor` as defined under
            `self.processor` in
            [diffusers.cross_attention](https://github.com/huggingface/diffusers/blob/main/src/diffusers/models/cross_attention.py).
        guidance_rescale (`float`, *optional*, defaults to 0.7):
            Guidance rescale factor proposed by [Common Diffusion Noise Schedules and Sample Steps are
            Flawed](https://arxiv.org/pdf/2305.08891.pdf) `guidance_scale` is defined as `φ` in equation 16. of
            [Common Diffusion Noise Schedules and Sample Steps are Flawed](https://arxiv.org/pdf/2305.08891.pdf).
            Guidance rescale factor should fix overexposure when using zero terminal SNR.
        original_size (`Tuple[int]`, *optional*, defaults to (1024, 1024)):
            TODO
        crops_coords_top_left (`Tuple[int]`, *optional*, defaults to (0, 0)):
            TODO
        target_size (`Tuple[int]`, *optional*, defaults to (1024, 1024)):
            TODO
        aesthetic_score (`float`, *optional*, defaults to 6.0):
            TODO
        negative_aesthetic_score (`float`, *optional*, defaults to 2.5):
            TDOO


    """  # noqa

    strength: float = 0.3
    num_inference_steps: int = 50
    guidance_scale: float = 7.5
    num_images_per_prompt: Optional[int] = 1
    eta: float = 0.0
    callback_steps: int = 1
    cross_attention_kwargs: Optional[Dict[str, Any]] = None
    guidance_rescale: float = 0.7
    original_size: Optional[Tuple[int, int]] = None
    crops_coords_top_left: Tuple[int, int] = (0, 0)
    target_size: Optional[Tuple[int, int]] = None
    aesthetic_score: float = 6.0
    negative_aesthetic_score: float = 2.5


# Encapsulating data classes for input -------------------------------------------------

# These have the following 4 attributes:
# - input: these are inputs sent to the inference task
# - inference_options: these are options that affect the run
# - output: these are additional arguments that specify output behavior


@dataclass
class StableDiffusionPipelineInput:
    input: InputForStableDiffusionPipeline
    inference_options: InferenceOptionsForStableDiffusionPipeline
    output: OutputForStableDiffusionPipeline


@dataclass
class StableDiffusionXLPipelineInput:
    input: InputForStableDiffusionXLPipeline
    inference_options: InferenceOptionsForStableDiffusionXLPipeline
    output: OutputForStableDiffusionPipeline


@dataclass
class StableDiffusionXLImg2ImgPipelineInput:
    input: InputForStableDiffusionXLImg2ImgPipeline
    inference_options: InferenceOptionsForStableDiffusionXLImg2ImgPipeline
    output: OutputForStableDiffusionPipeline



import click

from ..utils.printer import print_result


@click.group()
@click.pass_context
def cli(context):
    """Hardware"""
    pass


@cli.command("systeminfo")
@click.pass_context
def systeminfo_command(context):
    """Get System Info"""
    spice = context.obj.get("SPICE")
    message = spice.hardware.get_system_info()
    print_result(message=message, context=context)


@cli.command("register")
@click.pass_context
def register_command(context):
    """Register Hardware with Spice"""
    spice = context.obj.get("SPICE")
    result = spice.hardware.register()
    if result.get("registerHardware"):
        message = "Hardware registered successfully"
    print_result(message=message, context=context, fg="green")


@cli.command("checkin")
@click.pass_context
def checkin_command(context):
    """Checkin Hardware with Spice"""
    spice = context.obj.get("SPICE")
    result = spice.hardware.check_in_http()
    if result.get("checkIn"):
        message = "Hardware checked in successfully"
    print_result(message=message, context=context, fg="green")



import json
import logging
import os
from compel import Compel, ReturnedEmbeddingsType
from pathlib import Path
from typing import Optional, Dict, Any, Type, Union
from dataclasses import asdict

from diffusers import (
    DiffusionPipeline,
    StableDiffusionPipeline,
    StableDiffusionXLPipeline,
    StableDiffusionXLImg2ImgPipeline,
)
from gql import gql
from gql.transport.exceptions import TransportQueryError

from spice_agent.inference.types import (
    OutputForStableDiffusionPipeline,
    InputForStableDiffusionPipeline,
    InputForStableDiffusionXLPipeline,
    InputForStableDiffusionXLImg2ImgPipeline,
    InferenceOptionsForStableDiffusionPipeline,
    InferenceOptionsForStableDiffusionXLPipeline,
    InferenceOptionsForStableDiffusionXLImg2ImgPipeline,
    StableDiffusionPipelineInput,
    StableDiffusionXLPipelineInput,
    StableDiffusionXLImg2ImgPipelineInput,
)
from spice_agent.utils.config import SPICE_INFERENCE_DIRECTORY

# from torch.mps import empty_cache as mps_empty_cache ## SAVE FOR LATER

os.environ["PYTORCH_ENABLE_MPS_FALLBACK"] = "1"

import torch  # noqa
import transformers  # noqa
from transformers.pipelines.base import PipelineException  # noqa

LOGGER = logging.getLogger(__name__)


def _init_compel(
    device, pipeline: DiffusionPipeline, prompt: str, negative_prompt: str = ""
) -> Compel:
    """
    Initializes compel text embedding system based on type of pipeline.
    """

    # Compel settings consistent across compel objects
    DEFAULT_COMPEL_ARGUMENTS = {
        "truncate_long_prompts": False,
        "device": device,
    }

    # Currently, these are the only supported pipelines
    if not isinstance(
        pipeline,
        (
            StableDiffusionPipeline,
            StableDiffusionXLPipeline,
            StableDiffusionXLImg2ImgPipeline,
        ),
    ):
        message = (
            f"Prompt embeddings are not supported for pipeline {type(pipeline)}"  # noqa
        )
        LOGGER.warn(message)
        return None

    # Check if embeddings are necessary
    if pipeline.tokenizer:
        model_max_length = pipeline.tokenizer.model_max_length
        if len(prompt) <= model_max_length and len(negative_prompt) <= model_max_length:
            return None

    # Configure compel objects for StableDiffusion pipeline
    if isinstance(pipeline, StableDiffusionPipeline):
        return Compel(
            tokenizer=pipeline.tokenizer,
            text_encoder=pipeline.text_encoder,
            returned_embeddings_type=ReturnedEmbeddingsType.LAST_HIDDEN_STATES_NORMALIZED,
            requires_pooled=False,
            **DEFAULT_COMPEL_ARGUMENTS,
        )

    # Configure compel objects for StableDiffusionXL pipelines
    if isinstance(pipeline, StableDiffusionXLPipeline):
        return Compel(
            tokenizer=[pipeline.tokenizer, pipeline.tokenizer_2],
            text_encoder=[pipeline.text_encoder, pipeline.text_encoder_2],
            returned_embeddings_type=ReturnedEmbeddingsType.PENULTIMATE_HIDDEN_STATES_NON_NORMALIZED,
            requires_pooled=[False, True],
            **DEFAULT_COMPEL_ARGUMENTS,
        )

    if isinstance(pipeline, StableDiffusionXLImg2ImgPipeline):
        return Compel(
            tokenizer=pipeline.tokenizer_2,
            text_encoder=pipeline.text_encoder_2,
            returned_embeddings_type=ReturnedEmbeddingsType.PENULTIMATE_HIDDEN_STATES_NON_NORMALIZED,
            requires_pooled=True,
            **DEFAULT_COMPEL_ARGUMENTS,
        )

    return None


def _run_compel(
    compel: Compel, pipeline: DiffusionPipeline, prompt: str, negative_prompt: str = ""
) -> InputForStableDiffusionPipeline | InputForStableDiffusionXLPipeline | None:
    embeddings = None
    prompt_embeds = None
    pooled_prompt_embeds = None
    negative_prompt_embeds = None
    negative_pooled_prompt_embeds = None

    if compel.requires_pooled:
        # Get pooled embeddings
        prompt_embeds, pooled_prompt_embeds = compel(prompt)
        negative_prompt_embeds, negative_pooled_prompt_embeds = compel(negative_prompt)
    else:
        prompt_embeds = compel(prompt)
        negative_prompt_embeds = compel(negative_prompt)

    # Pad prompt embeds
    [
        padded_prompt_embeds,
        padded_negative_prompt_embeds,
    ] = compel.pad_conditioning_tensors_to_same_length(
        [prompt_embeds, negative_prompt_embeds]
    )

    if isinstance(pipeline, StableDiffusionPipeline):
        embeddings = InputForStableDiffusionPipeline(
            prompt_embeds=padded_prompt_embeds,
            negative_prompt_embeds=padded_negative_prompt_embeds,
        )
    elif isinstance(
        pipeline, (StableDiffusionXLPipeline, StableDiffusionXLImg2ImgPipeline)
    ):
        embeddings = InputForStableDiffusionXLPipeline(
            prompt_embeds=padded_prompt_embeds,
            negative_prompt_embeds=padded_negative_prompt_embeds,
            pooled_prompt_embeds=pooled_prompt_embeds,
            negative_pooled_prompt_embeds=negative_pooled_prompt_embeds,
        )

    return embeddings


def get_input_for_stable_diffusion_pipeline(
    device, pipeline: DiffusionPipeline, prompt: str, negative_prompt: str = ""
) -> InputForStableDiffusionPipeline | InputForStableDiffusionXLPipeline:
    """
    Returns prompt options for embeddings if a compel object exists; otherwise,
    prompt options without embeddings are returned
    """

    LOGGER.info(""" [*] Generating embeddings. (Ignore token indices complaint)""")

    without_embeddings = InputForStableDiffusionPipeline(
        prompt=prompt, negative_prompt=negative_prompt
    )
    compel = _init_compel(device, pipeline, prompt, negative_prompt)

    if compel:
        try:
            embeddings = _run_compel(compel, pipeline, prompt, negative_prompt)
            if not embeddings:
                return without_embeddings
        except Exception as exception:
            LOGGER.error(
                f""" [*] Embedding generation failed with exception {exception}! Falling back to normal prompt."""  # noqa
            )

    return without_embeddings


class Inference:
    def __init__(self, spice) -> None:
        self.spice = spice
        self.device = self.spice.get_device()
        self.pipe = None
        self.inference_job_id = None
        self.pipe_input: Optional[
            Union[
                StableDiffusionPipelineInput,
                StableDiffusionXLPipelineInput,
                StableDiffusionXLImg2ImgPipelineInput,
            ]
        ] = None

        # logging.basicConfig(level=logging.INFO)
        if self.spice.DEBUG:
            transformers.logging.set_verbosity_debug()  # type: ignore
            logging.getLogger("transformers.modeling_utils").setLevel(logging.ERROR)
            logging.getLogger("pika").setLevel(logging.ERROR)

    def _update_inference_job(
        self,
        inference_job_id: str,
        status: Optional[str] = None,
        status_details: Optional[Dict[str, Any]] = None,
        was_guarded: Optional[bool] = None,
        text_output: Optional[str] = None,
        file_outputs_ids: list[str] = [],
    ):
        mutation = gql(
            """
            mutation updateInferenceJob($input: UpdateInferenceJobInput!) {
                updateInferenceJob(input: $input) {
                    ... on InferenceJob {   
                        id
                        pk
                        createdAt
                        updatedAt
                        completedAt
                        status
                        statusDetails
                        model {
                            name
                            slug
                            repoId
                            isTextInput
                            isTextOutput
                            isFileInput
                            isFileOutput
                        }
                        textInput
                        textOutput
                        wasGuarded
                        options
                    }
                }
            }
            """
        )

        input: Dict[str, Any] = {"inferenceJobId": inference_job_id}
        if status is not None:
            input["status"] = status
        if status_details is not None:
            input["statusDetails"] = status_details
        if was_guarded is not None:
            input["wasGuarded"] = was_guarded
        if text_output is not None:
            input["textOutput"] = text_output
        if file_outputs_ids is not None or len(file_outputs_ids) > 0:
            input["fileOutputsIds"] = file_outputs_ids

        variables = {"input": input}

        try:
            result = self.spice.session.execute(mutation, variable_values=variables)
            return result
        except TransportQueryError as exception:
            if exception.errors:
                for error in exception.errors:
                    if error.get("message", "") == "Inference Job not found.":
                        LOGGER.error(
                            f""" [*] Inference Job ID: {inference_job_id} not found.\
                                  Exiting early."""
                        )
                        return None
                raise exception
            else:
                raise exception

    def _get_filtered_options(
        self,
        options: Dict[str, Any],
        inference_options: Type[
            Union[
                InferenceOptionsForStableDiffusionPipeline,
                InferenceOptionsForStableDiffusionXLPipeline,
                InferenceOptionsForStableDiffusionXLImg2ImgPipeline,
            ]
        ],
    ) -> Dict[str, Any]:
        return {
            key: value
            for key, value in options.items()
            if key in inference_options.__annotations__
        }

    def _get_inference_options_for_stable_diffusion(
        self, pipeline: DiffusionPipeline, options: Dict[str, Any]
    ) -> Union[
        InferenceOptionsForStableDiffusionPipeline,
        InferenceOptionsForStableDiffusionXLPipeline,
        InferenceOptionsForStableDiffusionXLImg2ImgPipeline,
    ]:
        """
        Parses any inference options that may be defined for
        a stable diffusion pipeline.
        """

        if isinstance(pipeline, StableDiffusionPipeline):
            filtered_options = self._get_filtered_options(
                options, InferenceOptionsForStableDiffusionPipeline
            )
            return InferenceOptionsForStableDiffusionPipeline(**filtered_options)
        elif isinstance(pipeline, StableDiffusionXLPipeline):
            filtered_options = self._get_filtered_options(
                options, InferenceOptionsForStableDiffusionXLPipeline
            )
            return InferenceOptionsForStableDiffusionXLPipeline(**filtered_options)
        elif isinstance(pipeline, StableDiffusionXLImg2ImgPipeline):
            filtered_options = self._get_filtered_options(
                options, InferenceOptionsForStableDiffusionXLImg2ImgPipeline
            )
            return InferenceOptionsForStableDiffusionXLImg2ImgPipeline(
                **filtered_options
            )
        else:
            raise ValueError(
                f"Pipeline {type(pipeline).__name__} has no supported inference options!"  # noqa
            )

    def _get_generator(self, seed: int) -> torch.Generator:
        # Note, completely reproducible results are not guaranteed across
        # PyTorch releases.
        return torch.manual_seed(seed)

    def callback_for_stable_diffusion(
        self, step: int, timestep: int, latents: torch.FloatTensor
    ) -> None:
        """
        A function that will be called every `callback_steps` steps during inference. The function will be
        called with the following arguments: `callback(step: int, timestep: int, latents: torch.FloatTensor)`.
        """  # noqa

        # Need access to vae to decode here:
        if self.pipe and self.inference_job_id and self.pipe_input:
            # Update progress on backend
            self._update_inference_job(
                inference_job_id=self.inference_job_id,
                status_details={
                    "progress": (
                        (step + 1)
                        / self.pipe_input.inference_options.num_inference_steps
                    )
                    * 100
                },
            )

    def run_pipeline(
        self,
        inference_job_id: str,
    ):  # noqa
        LOGGER.info(f""" [*] Inference Job ID: {inference_job_id}.""")

        result = self._update_inference_job(
            inference_job_id=inference_job_id,
            status="RUNNING",
        )
        if not result:
            LOGGER.error(
                f""" [*] Inference Job ID: {inference_job_id} not found.\
                                  Exiting early."""
            )
            return

        inference_job_id = result["updateInferenceJob"]["id"]
        self.inference_job_id = inference_job_id

        model_repo_id = result["updateInferenceJob"]["model"]["repoId"]
        text_input = result["updateInferenceJob"]["textInput"]
        is_text_input = result["updateInferenceJob"]["model"]["isTextInput"]
        is_text_output = result["updateInferenceJob"]["model"]["isTextOutput"]
        result["updateInferenceJob"]["model"]["isFileInput"]
        is_file_output = result["updateInferenceJob"]["model"]["isFileOutput"]
        options = result["updateInferenceJob"]["options"]

        LOGGER.info(f""" [*] Model: {model_repo_id}.""")
        LOGGER.info(f""" [*] Text Input: '{text_input}'""")

        variant = "fp16"
        torch_dtype = torch.float16
        if torch.backends.mps.is_available():
            variant = "fp32"
            torch_dtype = torch.float32
        #     mps_empty_cache()
        # if torch.cuda.is_available():
        #     torch.cuda.empty_cache()

        response = None
        try:
            if is_text_input and is_text_output:
                result = None
                pipe = transformers.pipeline(model=model_repo_id, device=self.device)
                result = pipe(text_input)

                response = self._update_inference_job(
                    inference_job_id=inference_job_id,
                    status="COMPLETE",
                    text_output=json.dumps(result),
                )
            elif is_text_input and is_file_output:
                SPICE_INFERENCE_DIRECTORY.mkdir(parents=True, exist_ok=True)
                save_at = Path(SPICE_INFERENCE_DIRECTORY / f"{inference_job_id}.png")

                prompt = text_input
                negative_prompt = options.get("negative_prompt", "")
                generator = self._get_generator(int(options.get("seed", -1)))
                callback = self.callback_for_stable_diffusion

                was_guarded = False
                if not save_at.exists():
                    pipe = DiffusionPipeline.from_pretrained(
                        model_repo_id,
                        torch_dtype=torch_dtype,
                        variant=variant,
                        use_safetensors=True,
                    )
                    pipe = pipe.to(self.device)  # type: ignore
                    self.pipe = pipe

                    # Get input for Stable Diffusion Pipelines
                    input_for_stable_diffusion_pipeline = (
                        get_input_for_stable_diffusion_pipeline(
                            self.device, pipe, prompt, negative_prompt
                        )
                    )

                    # Configure Stable Diffusion TASK
                    if isinstance(pipe, StableDiffusionPipeline):
                        # Configure inference options for stable diffusion pipeline
                        inference_options_for_stable_diffusion = (
                            self._get_inference_options_for_stable_diffusion(
                                pipe, options
                            )
                        )

                        # Configure output for stable diffusion pipeline
                        output_for_stable_diffusion_pipeline = OutputForStableDiffusionPipeline(  # noqa
                            # generator=generator,
                            return_dict=False,
                        )

                        if not isinstance(
                            input_for_stable_diffusion_pipeline,
                            InputForStableDiffusionPipeline,
                        ):
                            raise ValueError(
                                "Input for stable diffusion pipeline not configured!"
                            )

                        if not isinstance(
                            inference_options_for_stable_diffusion,
                            InferenceOptionsForStableDiffusionPipeline,
                        ):
                            raise ValueError(
                                "Inference options for stable diffusion pipeline not configured!"  # noqa
                            )

                        # Specify input for stable diffusion pipeline
                        stable_diffusion_pipeline_input = StableDiffusionPipelineInput(
                            input=input_for_stable_diffusion_pipeline,
                            inference_options=inference_options_for_stable_diffusion,
                            output=output_for_stable_diffusion_pipeline,
                        )
                        self.pipe_input = stable_diffusion_pipeline_input

                        pipe_result = pipe(
                            **asdict(stable_diffusion_pipeline_input.input),
                            **asdict(stable_diffusion_pipeline_input.inference_options),
                            **asdict(stable_diffusion_pipeline_input.output),
                            generator=generator,
                            callback=callback,
                        )  # type:ignore
                    # Configure MOE for xl diffusion base + refinement TASK
                    elif isinstance(pipe, StableDiffusionXLPipeline):
                        # Configure input for stable diffusion xl pipeline
                        input_for_stable_diffusion_xl = (
                            InputForStableDiffusionXLPipeline(
                                **asdict(input_for_stable_diffusion_pipeline)
                            )
                        )

                        # Configure inference options for stable diffusion xl pipeline
                        inference_options_for_stable_diffusion_xl = (
                            self._get_inference_options_for_stable_diffusion(
                                pipe, options
                            )
                        )

                        if not isinstance(
                            inference_options_for_stable_diffusion_xl,
                            InferenceOptionsForStableDiffusionXLPipeline,
                        ):
                            raise ValueError(
                                f"Inference options are incorrectly configured for {StableDiffusionPipeline.__name__}"  # noqa
                            )

                        # Configure output for stable diffusion xl pipeline
                        output_for_stable_diffusion_xl_pipeline = OutputForStableDiffusionPipeline(  # noqa
                            output_type="latent",
                            # generator=generator,
                        )

                        # Specify input for stable diffusion xl pipeline
                        stable_diffusion_pipeline_xl_input = StableDiffusionXLPipelineInput(  # noqa
                            input=input_for_stable_diffusion_xl,
                            inference_options=inference_options_for_stable_diffusion_xl,
                            output=output_for_stable_diffusion_xl_pipeline,
                        )
                        self.pipe_input = stable_diffusion_pipeline_xl_input

                        latents = pipe(
                            **asdict(stable_diffusion_pipeline_xl_input.input),
                            **asdict(
                                stable_diffusion_pipeline_xl_input.inference_options
                            ),
                            **asdict(stable_diffusion_pipeline_xl_input.output),
                            generator=generator,
                            callback=callback,
                        ).images  # type: ignore

                        refiner = StableDiffusionXLImg2ImgPipeline.from_pretrained(
                            "stabilityai/stable-diffusion-xl-refiner-1.0",
                            text_encoder_2=pipe.text_encoder_2,
                            vae=pipe.vae,
                            torch_dtype=torch_dtype,
                            variant=variant,
                            use_safetensors=True,
                        )
                        refiner = refiner.to(self.device)

                        # Configure input for stable diffusion xl img2img pipeline
                        input_for_stable_diffusion_xl_img2img = (
                            InputForStableDiffusionXLImg2ImgPipeline(
                                image=latents, **asdict(input_for_stable_diffusion_xl)
                            )
                        )

                        # Configure inference options for stable diffusion xl img2img pipeline # noqa
                        inference_options_for_stable_diffusion_xl_img2img = (
                            self._get_inference_options_for_stable_diffusion(
                                refiner, options
                            )
                        )

                        if not isinstance(
                            inference_options_for_stable_diffusion_xl_img2img,
                            InferenceOptionsForStableDiffusionXLImg2ImgPipeline,
                        ):
                            raise ValueError(
                                f"Inference options are incorrectly configured for {StableDiffusionXLImg2ImgPipeline.__name__}"  # noqa
                            )

                        # Configure output for stable diffusion xl pipeline
                        output_for_stable_diffusion_xl_img2img_pipeline = OutputForStableDiffusionPipeline(  # noqa
                            # generator=generator,
                        )

                        # Specify input for stable diffusion xl pipeline
                        stable_diffusion_pipeline_xl_img2img_input = StableDiffusionXLImg2ImgPipelineInput(  # noqa
                            input=input_for_stable_diffusion_xl_img2img,
                            inference_options=inference_options_for_stable_diffusion_xl_img2img,
                            output=output_for_stable_diffusion_xl_img2img_pipeline,
                        )
                        self.pipe_input = stable_diffusion_pipeline_xl_img2img_input

                        # Note, we do not attach a callback here since refinement
                        # of the image is not as time consuming
                        pipe_result = refiner(
                            **asdict(stable_diffusion_pipeline_xl_img2img_input.input),
                            **asdict(
                                stable_diffusion_pipeline_xl_img2img_input.inference_options
                            ),
                            **asdict(stable_diffusion_pipeline_xl_img2img_input.output),
                            generator=generator,
                        )  # type: ignore
                    else:
                        raise ValueError(
                            f"Pipeline {type(pipe).__name__} has no supported inference options!"  # noqa
                        )

                    # pipe returns a tuple in the form the first element is a list with
                    # the generated images, and the second element is a list of `bool`s
                    # denoting whether the corresponding generated image likely
                    # represents "not-safe-for-work" (nsfw) content, according to the
                    # `safety_checker`.
                    # TODO decode output based on output of actual pipeline
                    result = pipe_result[0][0]  # type: ignore
                    if len(pipe_result) > 1 and pipe_result[1]:  # type: ignore
                        was_guarded = pipe_result[1][0]  # type: ignore
                    result.save(save_at)  # type: ignore
                else:
                    LOGGER.info(f""" [*] File already exists at: {save_at}""")

                upload_file_response = self.spice.uploader.upload_file_via_api(
                    path=save_at
                )
                file_id = upload_file_response.json()["data"]["uploadFile"]["id"]
                response = self._update_inference_job(
                    inference_job_id=inference_job_id,
                    status="COMPLETE",
                    file_outputs_ids=file_id,
                    was_guarded=was_guarded,
                )
            LOGGER.info(f""" [*] COMPLETE. Result: " {result}""")
            return response
        except PipelineException as exception:
            message = f"""Input: "{input}" threw exception: {exception}"""
            LOGGER.error(message)
            self._update_inference_job(
                inference_job_id=inference_job_id,
                status="ERROR",
                text_output=json.dumps(message),
            )



from gql import gql

from ..utils.config import read_config_file, update_config_file


class Auth:
    def __init__(self, spice) -> None:
        self.spice = spice

    def whoami(self):
        query = gql(
            """
            query whoami {
                whoami {
                    username
                }
            }
        """
        )
        result = self.spice.session.execute(query)
        return result

    def setup_config(
        self,
        username: str,
        token: str,
        host: str = "api.spice.cloud",
        transport: str = "https",
    ):
        full_config = read_config_file()
        new_host_config = {
            "username": username,
            "token": token,
            "transport": transport,
        }

        if existing_host_config := full_config.get(host, None):
            full_config[host] = {**existing_host_config, **new_host_config}
        else:
            full_config[host] = new_host_config
        update_config_file(new_config=full_config)



import click

from ..utils.printer import print_result


@click.group()
@click.pass_context
def cli(context):
    """Inference"""
    pass


@cli.command("verify-torch")
@click.pass_context
def verify_torch_command(context):
    """Get System Info"""
    spice = context.obj.get("SPICE")
    message = spice.inference.verify_torch()
    print_result(message=message, context=context, fg="green")


@cli.command("run")
@click.option(
    "--model",
    required=False,
    default="bert-base-uncased",
    show_default="bert-base-uncased",
    help="Model from HuggingFace",
)
@click.option(
    "--input",
    required=False,
    default="spice.cloud is [MASK]!",
    show_default="spice.cloud is [MASK]!",
    help="Input for model to run inference",
)
@click.pass_context
def run_command(context, model: str, input: str):
    """Run HuggingFace Pipeline with Model from hub and a basic input."""
    spice = context.obj.get("SPICE")
    message = spice.inference.run_pipeline(model=model, input=input)
    print_result(message=message, context=context)



import click

from ..utils.printer import print_result


@click.group()
@click.pass_context
def cli(context):
    """Daemon"""
    pass


@cli.command("install")
@click.pass_context
def install_daemon_command(context):
    """Install the full spice Daemon"""
    spice = context.obj.get("SPICE")
    result = spice.daemons.install()
    print_result(message=result, context=context)


@cli.command("uninstall")
@click.pass_context
def uninstall_daemon_command(context):
    """Uninstall the full spice Daemon"""
    spice = context.obj.get("SPICE")
    result = spice.daemons.uninstall()
    print_result(message=result, context=context)


@cli.command("stop")
@click.pass_context
def stop_daemon_command(context):
    """Stop spice Daemon"""
    spice = context.obj.get("SPICE")
    result = spice.daemons.stop()
    message = "stopping spice daemon"
    if context.obj["JSON"]:
        message = result
    print_result(message=message, context=context)


@cli.command("start")
@click.pass_context
def start_daemon_command(context):
    """Start spice Daemon"""
    spice = context.obj.get("SPICE")
    result = spice.daemons.start()
    message = "starting spice daemon"
    if context.obj["JSON"]:
        message = result
    print_result(message=message, context=context)


@cli.command("logs")
@click.pass_context
def log_daemon_command(context):
    """Logs from spice Daemon"""
    spice = context.obj.get("SPICE")
    result = spice.daemons.logs()
    print_result(message=result, context=context)



import json
import logging
import os
import ssl
import sys

import pika
from pika.exceptions import AMQPConnectionError, ConnectionClosedByBroker
from retry import retry

from spice_agent.utils.config import (
    SPICE_ROUND_VERIFICATION_FILEPATH,
    SPICE_TRAINING_FILEPATH,
    read_config_file,
)
from spice_agent.utils.version import get_current_version

os.environ["PYTORCH_ENABLE_MPS_FALLBACK"] = "1"

import transformers  # noqa

LOGGER = logging.getLogger(__name__)

ACTIVE_VERIFY_ROUND_STATUSES = [
    "CLAIMED",
    "DOWNLOADING_ROUND_MODEL",
    "DOWNLOADING_VERIFICATION_DATASET",
    "VERIFYING",
]
ACTIVE_TRAINING_ROUND_STEP_STATUSES = [
    "CLAIMED",
    "DOWNLOADING_MODEL",
    "DOWNLOADING_DATASET",
    "TRAINING",
    "UPLOADING_STEP_MODEL",
]
ACTIVE_TESTING_STEP_STATUSES = [
    "TRAINING_COMPLETE",
    "DOWNLOADING_TESTING_DATASET",
    "TESTING",
]
ACTIVE_UPLOADING_STEP_STATUSES = [
    "TESTING_COMPLETE",
    "REQUEST_UPLOAD",
    "UPLOADING",
]

ACTIVE_STATUSES = (
    ACTIVE_VERIFY_ROUND_STATUSES
    + ACTIVE_TRAINING_ROUND_STEP_STATUSES
    + ACTIVE_TESTING_STEP_STATUSES
    + ACTIVE_UPLOADING_STEP_STATUSES
)  # noqa


class Worker:
    def __init__(self, spice) -> None:
        self.spice = spice
        self.channel = None
        self.device = self.spice.get_device()
        self.can_do_validation = "cuda" in self.device.type
        self.ACTIVE_STATUSES = ACTIVE_STATUSES

        # logging.basicConfig(level=logging.INFO)
        if self.spice.DEBUG:
            transformers.logging.set_verbosity_debug()
            logging.getLogger("transformers.modeling_utils").setLevel(logging.ERROR)
            logging.getLogger("pika").setLevel(logging.ERROR)

    def _claim_message_callback(self, channel, method, properties, body):
        """
        Based on if we get a run_id, training_round_step_id, training_round_id from the
        message consume
        """
        LOGGER.info(" [*] Processing message.")
        data = json.loads(body.decode("utf-8"))

        inference_job_id = data.get("inference_job_id", None)
        training_round_step_id = data.get("training_round_step_id", None)
        training_round_id = data.get("training_round_id", None)

        if (
            not inference_job_id
            and not training_round_step_id
            and not training_round_id
        ):
            raise Exception(
                f'No inference_job_id or training_round_step_id or training_round_id found in message body: {body.decode("utf-8")}'  # noqa
            )

        message_acked = False
        if inference_job_id:
            self.spice.inference._update_inference_job(
                inference_job_id=inference_job_id, status="CLAIMED"
            )
            # ack message at inference level so another machine does not steal the
            # message while inference is running
            channel.basic_ack(delivery_tag=method.delivery_tag)
            message_acked = True
            self.spice.inference.run_pipeline(inference_job_id=inference_job_id)
            LOGGER.info(" [*] Completed inference job.")

        if training_round_id:
            result = self.spice.training._update_training_round(
                training_round_id=training_round_id, status="CLAIMED"
            )
            if result is not None:
                LOGGER.info(" [*] Obtained training round.")
                LOGGER.info(" [*] Stopping consumer, to begin training round.")
                if channel:
                    channel.basic_ack(delivery_tag=method.delivery_tag)
                    message_acked = True
                    channel.stop_consuming()
                    channel.close()

        if training_round_step_id:
            result = self.spice.training._update_training_round_step(
                training_round_step_id=training_round_step_id, status="CLAIMED"
            )
            if result is not None:
                LOGGER.info(" [*] Obtained training round step.")
                LOGGER.info(" [*] Stopping consumer, to begin training round step.")
                if channel:
                    channel.basic_ack(delivery_tag=method.delivery_tag)
                    message_acked = True
                    channel.stop_consuming()
                    channel.close()

        if not message_acked and channel.is_open:
            channel.basic_ack(delivery_tag=method.delivery_tag)
        elif not message_acked and not channel.is_open:
            LOGGER.info(" [*] Channel closed already. Cannot ack message.")
        else:
            pass

    def _create_channel(self):
        credentials = pika.PlainCredentials(
            self.spice.host_config["fingerprint"],
            self.spice.host_config["rabbitmq_password"],
        )

        host = self.spice.host_config["rabbitmq_host"]
        ssl_options = None
        if "localhost" not in host:
            context = ssl.SSLContext(ssl.PROTOCOL_TLSv1_2)
            context.verify_mode = ssl.CERT_NONE
            ssl_options = pika.SSLOptions(context)

        connection = pika.BlockingConnection(
            pika.ConnectionParameters(
                host=self.spice.host_config["rabbitmq_host"],
                port=self.spice.host_config["rabbitmq_port"],
                virtual_host="agent",
                credentials=credentials,
                ssl_options=ssl_options,
            )
        )

        self.channel = connection.channel()
        LOGGER.info(" [*] Channel connected.")

    @retry(AMQPConnectionError, delay=0, jitter=(1, 3))
    def _consume(self):
        if self.channel is None or self.channel.is_closed:
            self._create_channel()

        if self.channel is None:
            raise Exception(" [*] Channel cannot be opened.")

        self.channel.queue_declare(
            queue=self.spice.host_config["fingerprint"],
            exclusive=True,
            durable=True,
            auto_delete=True,
        )

        self.channel.queue_bind(
            exchange="amq.headers",
            queue=self.spice.host_config["fingerprint"],
            arguments={
                "x-match": "all",
                "fingerprint": self.spice.host_config["fingerprint"],
            },
        )

        self.spice.hardware.check_in_http(is_available=True, is_online=True)

        self.channel.basic_consume(
            queue=self.spice.host_config["fingerprint"],
            exclusive=True,
            on_message_callback=self._claim_message_callback,
        )

        try:
            LOGGER.info(" [*] Waiting for messages. To exit press CTRL+C")
            self.channel.start_consuming()
        except KeyboardInterrupt:
            LOGGER.info(" [*] Stopping RabbitMQ consumer...")
            if self.channel:
                self.channel.stop_consuming()
                self.channel.close()
                self.channel = None
            self.spice.hardware.check_in_http(is_available=False, is_online=False)
            sys.exit()
        except ConnectionClosedByBroker:
            LOGGER.info(" [*] Connection closed by Broker.")
            if self.channel:
                self.channel.stop_consuming()
                self.channel.close()
                self.channel = None
            self.spice.hardware.check_in_http(is_available=False, is_online=False)
        except Exception as exception:
            LOGGER.info(f"Exception: {exception}")
            if self.channel:
                self.channel.stop_consuming()
                self.channel.close()
                self.channel = None
            self.spice.hardware.check_in_http(is_available=False, is_online=False)
            raise exception

    def start(self):
        LOGGER.info(" [*] ✨ spice worker")
        LOGGER.info(f" [*] Version: {get_current_version()}")
        self.spice.hardware.update_hardware_system_info()
        self.spice.hardware.check_in_http(is_available=False, is_online=True)
        try:
            while True:
                # check if this machine picked up a training round already
                training_config = read_config_file(filepath=SPICE_TRAINING_FILEPATH)

                has_step_to_train = False
                has_step_to_test = False
                has_model_to_upload = False
                has_round_to_verify = False

                if training_config.get("id", None) is not None:
                    # check that training round step is active
                    step_status = training_config.get("status", None)
                    if step_status in ACTIVE_TRAINING_ROUND_STEP_STATUSES:
                        has_step_to_train = True
                    if step_status in ACTIVE_TESTING_STEP_STATUSES:
                        has_step_to_test = True
                    elif step_status in ACTIVE_UPLOADING_STEP_STATUSES:
                        has_model_to_upload = True

                if self.can_do_validation:
                    verification_config = read_config_file(
                        filepath=SPICE_ROUND_VERIFICATION_FILEPATH
                    )
                    if verification_config.get("id", None) is not None:
                        step_status = verification_config.get("status", None)
                        if step_status in ACTIVE_VERIFY_ROUND_STATUSES:
                            has_round_to_verify = True

                if (
                    not has_step_to_train
                    and not has_step_to_test
                    and not has_model_to_upload
                    and not has_round_to_verify
                ):
                    self._consume()

                if has_round_to_verify:
                    LOGGER.info(
                        f" [*] Validating - Using Device: {self.device} for validation"
                    )
                    self.spice.training.verify_model()
                if has_step_to_train:
                    LOGGER.info(
                        f" [*] Training - Using Device: {self.device} for training"
                    )
                    self.spice.training.train_model()
                if has_step_to_test:
                    LOGGER.info(
                        f" [*] Testing - Using Device: {self.device} for testing"
                    )
                    self.spice.training.test_model()
                if has_model_to_upload:
                    LOGGER.info(" [*] Uploading")
                    self.spice.training.upload_models()

        except KeyboardInterrupt:
            LOGGER.info(" [*] Stopping worker...")
            if self.channel:
                self.channel.stop_consuming()
                self.channel.close()
                self.channel = None
            self.spice.hardware.check_in_http(is_available=False, is_online=False)
            sys.exit()
